from contextlib import contextmanager
from queue import Queue
from threading import Event
from typing import Any, Generator, Iterable, Optional, Type

import sounddevice as sd

from ...._model import AudioState
from ..._player import AudioPlayer
from ._generator import AudioGenerator
from ._mix import Mix
from ._output import AudioOutputCallback
from ._sound import Sound


class AudioSynthPlayer(AudioPlayer):
    """
    The ``AudioSynthPlayer`` can play synthetic sounds (specified either by MIDI
    note or raw frequency) to an audio device.
    """

    def __init__(
        self,
        *args,
        volume: float,
        channels: int,
        dtype,  # : DTypeLike,
        sounds: Optional[Iterable[Sound]] = None,
        **kwargs
    ):
        sounds = sounds or []
        self.mix = Mix(*sounds, volume=volume, channels=channels, dtype=dtype)

        super().__init__(*args, volume=volume, channels=channels, dtype=dtype, **kwargs)
        self._generator_stopped = Event()
        self._completed_callback_event = Event()
        self._audio_queue = Queue(  # Queue[NDArray[np.number]]
            maxsize=self.queue_size or 0
        )

    @property
    def _stream_type(self) -> Type[sd.OutputStream]:
        return sd.OutputStream

    @property
    def _audio_converter_type(self) -> None:
        pass

    def __setattr__(self, __name: str, __value: Any):
        """
        Make sure that the relevant attributes are synchronized to the mix
        object upon set/update.
        """
        if __name == 'volume':
            # Propagate the volume changes to the mix object.
            self.mix.volume = __value
        return super().__setattr__(__name, __value)

    def _on_converter_timeout(self, *_, **__) -> bool:
        """
        Don't break the audio stream if the output converter failed
        """
        return True

    @property
    def _stream_args(self) -> dict:
        """
        Register an :class:`.AudioOutputCallback` to fill up the audio buffers.
        """
        return {
            'callback': AudioOutputCallback(
                audio_queue=self._audio_queue,
                channels=self.channels,
                blocksize=self.blocksize,
                queue_timeout=self._queue_timeout,
                should_stop=lambda: self.should_stop
                or self._generator_stopped.is_set(),
                is_paused=lambda: self.state == AudioState.PAUSED,
            ),
            'finished_callback': self._completed_callback_event.set,
            **super()._stream_args,
        }

    @property
    def _queue_timeout(self) -> float:
        """
        Estimated max read/write timeout on the audio queue.
        """
        return self.blocksize * (self.queue_size or 5) / self.sample_rate

    @contextmanager
    def _audio_generator(self) -> Generator[AudioGenerator, None, None]:
        stop_generator = Event()
        gen = AudioGenerator(
            audio_queue=self._audio_queue,
            mix=self.mix,
            blocksize=self.blocksize,
            sample_rate=self.sample_rate,
            queue_timeout=self._queue_timeout,
            should_stop=lambda: self.should_stop or stop_generator.is_set(),
            wait_running=self._wait_running,
            on_stop=self._on_stop,
        )

        self._generator_stopped.clear()
        gen.start()
        yield gen

        stop_generator.set()
        gen.join()

    def _on_stop(self):
        self._generator_stopped.set()
        self.notify_stop()


# vim:sw=4:ts=4:et:



from dataclasses import asdict, dataclass
from typing import Optional, Union, List, Tuple

from platypush.plugins.camera import CameraInfo, Camera


@dataclass
class PiCameraInfo(CameraInfo):
    sharpness: int = 0
    contrast: int = 0
    brightness: int = 50
    video_stabilization: bool = False
    iso: int = 0
    exposure_compensation: int = 0
    exposure_mode: str = 'auto'
    meter_mode: str = 'average'
    awb_mode: str = 'auto'
    image_effect: str = 'none'
    color_effects: Optional[Union[str, List[str]]] = None
    zoom: Tuple[float, float, float, float] = (0.0, 0.0, 1.0, 1.0)
    led_pin: Optional[int] = None

    def to_dict(self) -> dict:
        return {
            'sharpness': self.sharpness,
            'contrast': self.contrast,
            'brightness': self.brightness,
            'video_stabilization': self.video_stabilization,
            'iso': self.iso,
            'exposure_compensation': self.exposure_compensation,
            'exposure_mode': self.exposure_mode,
            'meter_mode': self.meter_mode,
            'awb_mode': self.awb_mode,
            'image_effect': self.image_effect,
            'color_effects': self.color_effects,
            'zoom': self.zoom,
            'led_pin': self.led_pin,
            **asdict(super()),
        }


class PiCamera(Camera):
    info: PiCameraInfo  # type: ignore


# vim:sw=4:ts=4:et:



from marshmallow import pre_load

from .._base import SystemBaseSchema


class ConnectionBaseSchema(SystemBaseSchema):
    """
    Base schema for connections.
    """

    @pre_load
    def pre_load(self, data, **_) -> dict:
        data = super().pre_load(data)
        addr_mapping = {
            'laddr': ('local_address', 'local_port'),
            'raddr': ('remote_address', 'remote_port'),
        }

        # Parse laddr/raddr attributes
        for ext_attr, (addr_attr, port_attr) in addr_mapping.items():
            value = data.pop(ext_attr, None)
            if not value:
                data[addr_attr] = data[port_attr] = None
            elif isinstance(value, tuple):
                data[addr_attr], data[port_attr] = value
            elif isinstance(value, str):
                data[addr_attr] = value
                data[port_attr] = None

        # Handle enum values
        for attr in ['type', 'family']:
            value = data.pop(attr, None)
            if value is not None:
                data[attr] = value.name

        if data.get('status') == 'NONE':
            data['status'] = None

        return data



from enum import Enum
import json
from typing import Optional

from tornado.web import stream_request_body
from platypush.context import get_plugin

from platypush.config import Config
from platypush.plugins.camera import Camera, CameraPlugin, StreamWriter
from platypush.utils import get_plugin_name_by_class

from .. import StreamingRoute


class RequestType(Enum):
    """
    Models the camera route request type (video or photo)
    """

    UNKNOWN = ''
    PHOTO = 'photo'
    VIDEO = 'video'


@stream_request_body
class CameraRoute(StreamingRoute):
    """
    Route for camera streams.
    """

    _redis_queue_prefix = f'_platypush/{Config.get("device_id") or ""}/camera'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._camera: Optional[Camera] = None
        self._request_type = RequestType.UNKNOWN
        self._extension: str = ''

    @classmethod
    def path(cls) -> str:
        return r"/camera/([a-zA-Z0-9_./]+)/([a-zA-Z0-9_]+)\.?([a-zA-Z0-9_]+)?"

    def _get_camera(self, plugin: str) -> CameraPlugin:
        plugin_name = f'camera.{plugin.replace("/", ".")}'
        p = get_plugin(plugin_name)
        assert p, f'No such plugin: {plugin_name}'
        return p

    def _get_frame(
        self, camera: Camera, timeout: Optional[float] = None
    ) -> Optional[bytes]:
        if camera.stream:
            with camera.stream.ready:
                camera.stream.ready.wait(timeout=timeout)
                return camera.stream.frame

        return None

    def send_frame(self, camera: Camera):
        frame = None
        for _ in range(camera.info.warmup_frames):
            frame = self._get_frame(camera)

        if frame:
            self.write(frame)
            self.flush()

    def _set_request_type_and_extension(self, route: str, extension: str):
        if route in {'photo', 'frame'}:
            self._request_type = RequestType.PHOTO
            if extension == 'jpg':
                extension = 'jpeg'
            self._extension = extension or 'jpeg'
        elif route in {'video', 'feed'}:
            self._request_type = RequestType.VIDEO
            self._extension = extension or 'mjpeg'

    def _get_args(self, kwargs: dict):
        kwargs = {k: v[0].decode() for k, v in kwargs.items() if k != 't'}
        for k, v in kwargs.items():
            if k == 'resolution':
                v = json.loads(f'[{v}]')
            else:
                try:
                    v = int(v)
                except (ValueError, TypeError):
                    try:
                        v = float(v)
                    except (ValueError, TypeError):
                        pass

            kwargs[k] = v

        return kwargs

    @classmethod
    def _get_redis_queue(cls, camera: CameraPlugin, *_, **__) -> str:
        plugin_name = get_plugin_name_by_class(camera.__class__)
        assert plugin_name, f'No such plugin: {plugin_name}'
        return '/'.join(
            [
                cls._redis_queue_prefix,
                plugin_name,
                *map(
                    str,
                    [camera.camera_info.device] if camera.camera_info.device else [],
                ),
            ]
        )

    def get(self, plugin: str, route: str, extension: str = '') -> None:
        self._set_request_type_and_extension(route, extension)
        if not (self._request_type and self._extension):
            self.write_error(404, 'Not Found')
            return

        stream_class = StreamWriter.get_class_by_name(self._extension)
        camera = self._get_camera(plugin)
        redis_queue = self._get_redis_queue(camera)
        self.set_header('Content-Type', stream_class.mimetype)

        with camera.open(
            stream=True,
            stream_format=self._extension,
            frames_dir=None,
            redis_queue=redis_queue,
            **self._get_args(self.request.arguments),
        ) as session:
            camera.start_camera(session)
            if self._request_type == RequestType.PHOTO:
                self.send_frame(session)
            elif self._request_type == RequestType.VIDEO:
                self.forward_stream(camera)

        self.finish()



from dataclasses import dataclass
from typing import Optional

from platypush.schemas.mopidy import MopidyTrackSchema

from ._exc import EmptyTrackException


@dataclass
class MopidyTrack:
    """
    Model for a Mopidy track.
    """

    uri: str
    artist: Optional[str] = None
    title: Optional[str] = None
    album: Optional[str] = None
    artist_uri: Optional[str] = None
    album_uri: Optional[str] = None
    time: Optional[float] = None
    playlist_pos: Optional[int] = None
    track_id: Optional[int] = None
    track_no: Optional[int] = None
    date: Optional[str] = None
    genre: Optional[str] = None
    type: str = 'track'

    @classmethod
    def parse(cls, track: dict) -> Optional["MopidyTrack"]:
        """
        Parse a Mopidy track from a dictionary received from the Mopidy API.
        """
        try:
            return cls(**MopidyTrackSchema().load(track))  # type: ignore
        except EmptyTrackException:
            return None

    def to_dict(self) -> dict:
        """
        Convert the Mopidy track to a dictionary.
        """
        return dict(MopidyTrackSchema().dump(self))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MopidyTrack):
            return False

        return (
            self.uri == other.uri
            and self.artist == other.artist
            and self.title == other.title
            and self.album == other.album
            and self.time == other.time
        )



import re
import sys
from typing import Any, Mapping, Optional, Union

from platypush.context import get_plugin
from platypush.message.event import Event, EventMatchResult
from platypush.plugins.assistant import AssistantPlugin
from platypush.utils import get_plugin_name_by_class


class AssistantEvent(Event):
    """Base class for assistant events"""

    def __init__(
        self, *args, assistant: Optional[Union[str, AssistantPlugin]] = None, **kwargs
    ):
        """
        :param assistant: Name of the assistant plugin that triggered the event.
        """
        assistant = assistant or kwargs.get('assistant')
        if assistant:
            kwargs['plugin'] = kwargs['_assistant'] = (
                assistant
                if isinstance(assistant, str)
                else get_plugin_name_by_class(assistant.__class__)
            )

        super().__init__(*args, **kwargs)

    @property
    def assistant(self) -> Optional[AssistantPlugin]:
        assistant = self.args.get('_assistant')
        if not assistant:
            return None

        return get_plugin(assistant)

    def as_dict(self):
        evt_dict = super().as_dict()
        evt_args = {**evt_dict['args']}
        assistant = evt_args.pop('_assistant', None)
        if assistant:
            evt_args['assistant'] = assistant

        return {
            **evt_dict,
            'args': evt_args,
        }


class ConversationStartEvent(AssistantEvent):
    """
    Event triggered when a new conversation starts
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ConversationEndEvent(AssistantEvent):
    """
    Event triggered when a conversation ends
    """

    def __init__(self, *args, with_follow_on_turn: bool = False, **kwargs):
        """
        :param with_follow_on_turn: Set to true if the conversation expects a
            user follow-up, false otherwise
        """
        super().__init__(*args, with_follow_on_turn=with_follow_on_turn, **kwargs)


class ConversationTimeoutEvent(ConversationEndEvent):
    """
    Event triggered when a conversation times out
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ResponseEvent(AssistantEvent):
    """
    Event triggered when a response is processed by the assistant
    """

    def __init__(
        self, *args, response_text: str, with_follow_on_turn: bool = False, **kwargs
    ):
        """
        :param response_text: Response text processed by the assistant
        :param with_follow_on_turn: Set to true if the conversation expects a
            user follow-up, false otherwise
        """
        super().__init__(
            *args,
            response_text=response_text,
            with_follow_on_turn=with_follow_on_turn,
            **kwargs,
        )


class ResponseEndEvent(ConversationEndEvent):
    """
    Event triggered when a response has been rendered on the assistant.
    """

    def __init__(
        self, *args, response_text: str, with_follow_on_turn: bool = False, **kwargs
    ):
        """
        :param response_text: Response text rendered on the assistant.
        :param with_follow_on_turn: Set to true if the conversation expects a
            user follow-up, false otherwise.
        """
        super().__init__(
            *args,
            response_text=response_text,
            with_follow_on_turn=with_follow_on_turn,
            **kwargs,
        )


class NoResponseEvent(ConversationEndEvent):
    """
    Event triggered when a conversation ends with no response
    """


class SpeechRecognizedEvent(AssistantEvent):
    """
    Event triggered when a speech is recognized
    """

    def __init__(self, *args, phrase: str, **kwargs):
        """
        :param phrase: Recognized user phrase
        """

        super().__init__(*args, phrase=phrase, **kwargs)
        self.recognized_phrase = phrase.strip().lower()

    def matches_condition(self, condition):
        """
        Overrides matches condition, and stops the conversation to prevent the
        default assistant response if the event matched some event hook condition.
        """

        result = super().matches_condition(condition)
        if (
            result.is_match
            and condition.args.get('phrase')
            and self.assistant
            and self.assistant.stop_conversation_on_speech_match
        ):
            self.assistant.stop_conversation()

        return result

    def _matches_argument(self, argname, condition_value, event_args, result):
        """
        Overrides the default `_matches_argument` method to allow partial
        phrase matches and text extraction.

        Example::

            event_args = {
                'phrase': 'Hey dude turn on the living room lights'
            }

        - `self._matches_argument(argname='phrase', condition_value='Turn on the ${lights} lights')`
          will return `EventMatchResult(is_match=True, parsed_args={ 'lights': 'living room' })`

        - `self._matches_argument(argname='phrase', condition_value='Turn off the ${lights} lights')`
          will return `EventMatchResult(is_match=False, parsed_args={})`

        """

        if event_args.get(argname) == condition_value:
            # In case of an exact match, return immediately
            result.is_match = True
            result.score = sys.maxsize
            return result

        parsed_args = {}
        event_tokens = re.split(r'\s+', event_args.get(argname, '').strip().lower())
        condition_tokens = re.split(r'\s+', condition_value.strip().lower())

        while event_tokens and condition_tokens:
            event_token = event_tokens[0]
            condition_token = condition_tokens[0]

            if event_token == condition_token:
                event_tokens.pop(0)
                condition_tokens.pop(0)
                result.score += 1.5
            elif re.search(condition_token, event_token):
                m = re.search(f'({condition_token})', event_token)
                if m and m.group(1):
                    event_tokens.pop(0)
                    result.score += 1.25

                condition_tokens.pop(0)
            else:
                m = re.match(r'[^\\]*\${(.+?)}', condition_token)
                if m:
                    argname = m.group(1)
                    if argname not in parsed_args:
                        parsed_args[argname] = event_token
                        result.score += 1.0
                    else:
                        parsed_args[argname] += ' ' + event_token

                    if (len(condition_tokens) == 1 and len(event_tokens) == 1) or (
                        len(event_tokens) > 1
                        and len(condition_tokens) > 1
                        and event_tokens[1] == condition_tokens[1]
                    ):
                        # Stop appending tokens to this argument, as the next
                        # condition will be satisfied as well
                        condition_tokens.pop(0)

                    event_tokens.pop(0)
                else:
                    result.score -= 1.0
                    event_tokens.pop(0)

        # It's a match if all the tokens in the condition string have been satisfied
        result.is_match = len(condition_tokens) == 0
        if result.is_match:
            result.parsed_args = parsed_args

        return result


class IntentRecognizedEvent(AssistantEvent):
    """
    Event triggered when an intent is matched by a speech command.
    """

    def __init__(
        self, *args, intent: str, slots: Optional[Mapping[str, Any]] = None, **kwargs
    ):
        """
        :param intent: The intent that has been matched.
        :param slots: The slots extracted from the intent, as a key-value mapping.
        """
        super().__init__(*args, intent=intent, slots=slots or {}, **kwargs)

    def matches_condition(self, condition):
        """
        Overrides matches condition, and stops the conversation to prevent the
        default assistant response if the event matched some event hook condition.
        """
        result = super().matches_condition(condition)
        if (
            result.is_match
            and self.assistant
            and self.assistant.stop_conversation_on_speech_match
        ):
            self.assistant.stop_conversation()

        return result

    def _matches_argument(
        self, argname, condition_value, event_args, result: EventMatchResult
    ):
        if argname != 'slots':
            return super()._matches_argument(
                argname, condition_value, event_args, result
            )

        event_slots = set(event_args.get(argname, {}).items())
        slots = set(self.args.get(argname, {}).items())

        # All the slots in the condition must be in the event
        if slots.difference(event_slots) == 0:
            result.is_match = True
            result.score += 1
        else:
            result.is_match = False
            result.score = 0

        return result


class HotwordDetectedEvent(AssistantEvent):
    """
    Event triggered when a custom hotword is detected
    """

    def __init__(self, *args, hotword: Optional[str] = None, **kwargs):
        """
        :param hotword: The detected user hotword.
        """
        super().__init__(*args, hotword=hotword, **kwargs)


class VolumeChangedEvent(AssistantEvent):
    """
    Event triggered when the volume of the assistant changes.
    """

    def __init__(self, *args, volume: float, **kwargs):
        super().__init__(*args, volume=volume, **kwargs)


class AlertStartedEvent(AssistantEvent):
    """
    Event triggered when an alert starts on the assistant
    """


class AlertEndEvent(AssistantEvent):
    """
    Event triggered when an alert ends on the assistant
    """


class AlarmStartedEvent(AlertStartedEvent):
    """
    Event triggered when an alarm starts on the assistant
    """


class AlarmEndEvent(AlertEndEvent):
    """
    Event triggered when an alarm ends on the assistant
    """


class TimerStartedEvent(AlertStartedEvent):
    """
    Event triggered when a timer starts on the assistant
    """


class TimerEndEvent(AlertEndEvent):
    """
    Event triggered when a timer ends on the assistant
    """


class MicMutedEvent(AssistantEvent):
    """
    Event triggered when the microphone is muted.
    """


class MicUnmutedEvent(AssistantEvent):
    """
    Event triggered when the microphone is muted.
    """


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from typing import Dict, List, Type

from platypush.common.sensors import Numeric
from platypush.entities.devices import Device
from platypush.entities.distance import DistanceSensor
from platypush.entities.humidity import HumiditySensor
from platypush.entities.pressure import PressureSensor
from platypush.entities.sensors import NumericSensor
from platypush.entities.temperature import TemperatureSensor
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


@dataclass
class SensorEntityMapping:
    """
    Maps the dict-like data returned by the plugin to corresponding sensor
    entities.
    """

    name: str
    unit: str
    entity_type: Type[NumericSensor]

    @property
    def id(self):
        """
        The standardized external ID for the entity.
        """
        return f'bme280:{self.name}'


_sensor_entity_mappings = {
    mapping.name: mapping
    for mapping in [
        SensorEntityMapping(
            name='temperature',
            unit='°C',
            entity_type=TemperatureSensor,
        ),
        SensorEntityMapping(
            name='humidity',
            unit='%',
            entity_type=HumiditySensor,
        ),
        SensorEntityMapping(
            name='pressure',
            unit='Pa',
            entity_type=PressureSensor,
        ),
        SensorEntityMapping(
            name='altitude',
            unit='m',
            entity_type=DistanceSensor,
        ),
    ]
}


# pylint: disable=too-many-ancestors
class SensorBme280Plugin(SensorPlugin):
    """
    Plugin to interact with a `BME280 <https://shop.pimoroni.com/products/bme280-breakout>`_ environment sensor for
    temperature, humidity and pressure measurements over I2C interface.
    """

    def __init__(self, port: int = 1, **kwargs):
        """
        :param port: I2C port. 0 = /dev/i2c-0 (port I2C0), 1 = /dev/i2c-1 (port I2C1)
        """

        super().__init__(**kwargs)
        self.port = port
        self._smbus = None
        self._device = None

    def _get_device(self):
        from smbus import SMBus
        from bme280 import BME280

        if self._device:
            return self._device

        self._smbus = SMBus(self.port)
        self._device = BME280(i2c_dev=self._smbus)
        return self._device

    @action
    def get_measurement(self, *_, **__):
        """
        :returns: dict. Example:

        .. code-block:: python

           output = {
               "temperature": 21.0,   # Celsius
               "pressure": 101555.08, # Pascals
               "humidity": 23.543,    # percentage
               "altitude": 15.703     # meters
           }

        """

        device = self._get_device()
        return {
            'temperature': device.get_temperature(),
            'pressure': device.get_pressure() * 100,
            'humidity': device.get_humidity(),
            'altitude': device.get_altitude(),
        }

    def transform_entities(self, entities: Dict[str, Numeric]) -> List[Device]:
        sensors = []
        for sensor, value in entities.items():
            if value is None:
                continue

            mapping = _sensor_entity_mappings[sensor]
            sensors.append(
                mapping.entity_type(
                    id=mapping.id,
                    name=mapping.name,
                    value=value,
                    unit=mapping.unit,
                )
            )

        if not sensors:
            return []

        return [
            Device(
                id='bme280',
                name='BME280 Sensor',
                children=sensors,
            )
        ]


# vim:sw=4:ts=4:et:



from asyncio import Event as AsyncEvent
from collections import defaultdict
from dataclasses import dataclass, field
from threading import Event
from typing import Any, Dict, Iterable, Optional, Set

import aioxmpp
import aioxmpp.im.p2p

from .._types import RoomInvite, UserInvite


@dataclass
class OccupantState:
    """
    Models the state of a room occupant.
    """

    user_id: str
    nick: Optional[str]
    affiliation: Optional[str]
    role: Optional[str]
    is_self: bool
    available: bool
    state: Optional[str]

    @classmethod
    def load(cls, occupant_dict: Dict[str, Any]) -> "OccupantState":
        return cls(
            user_id=occupant_dict["user_id"],
            nick=occupant_dict.get("nick"),
            affiliation=occupant_dict.get("affiliation"),
            role=occupant_dict.get("role"),
            is_self=occupant_dict["is_self"],
            available=occupant_dict["available"],
            state=occupant_dict.get("state"),
        )


@dataclass
class RoomState:
    """
    Models the state of a room.
    """

    room_id: str
    joined: bool
    state: Optional[str]
    nick: Optional[str]
    password: Optional[str]
    members: Dict[str, OccupantState] = field(default_factory=dict)

    @classmethod
    def load(cls, room_dict: Dict[str, Any]) -> "RoomState":
        return cls(
            room_id=room_dict["room_id"],
            joined=room_dict["joined"],
            state=room_dict.get("state"),
            nick=room_dict.get("nick"),
            password=room_dict.get("password"),
            members={
                user_id: OccupantState.load(member)
                for user_id, member in room_dict.get("members", {}).items()
            },
        )


@dataclass
class SerializedState:
    """
    Serialized snapshot of the XMPP state, which can be more easily
    serialized/deserialized to JSON.
    """

    users: Iterable[str] = field(default_factory=list)
    """List of users on the subscriptions/contacts list."""
    rooms: Dict[str, RoomState] = field(default_factory=dict)
    """List of rooms the user has joined."""
    room_invites: Iterable[str] = field(default_factory=list)
    """List of room invites, by room_id."""
    user_invites: Iterable[str] = field(default_factory=list)
    """List of user invites, by user_id."""

    @classmethod
    def load(cls, state: Dict[str, Any]) -> "SerializedState":
        return cls(
            users=state.get("users", []),
            rooms={
                room_id: RoomState.load(room)
                for room_id, room in state.get("rooms", {}).items()
            },
            room_invites=state.get("room_invites", []),
            user_invites=state.get("user_invites", []),
        )


@dataclass
class XmppState:
    """
    Models the state of the XMPP client.
    """

    rooms: Dict[str, aioxmpp.muc.service.Room] = field(default_factory=dict)
    conversations: Dict[str, aioxmpp.im.p2p.Conversation] = field(default_factory=dict)
    users: Set[str] = field(default_factory=set)
    room_invites: Dict[str, RoomInvite] = field(
        default_factory=lambda: defaultdict(RoomInvite)
    )
    user_invites: Dict[str, UserInvite] = field(
        default_factory=lambda: defaultdict(UserInvite)
    )
    disconnect_notified: Event = field(default_factory=Event)
    should_stop: AsyncEvent = field(default_factory=AsyncEvent)
    pending_rooms: Set[str] = field(default_factory=set)
    """Set of rooms that are currently being joined"""

    @staticmethod
    def _occupant_user_id(occupant: aioxmpp.muc.service.Occupant) -> str:
        return (
            str(occupant.direct_jid.replace(resource=None))
            if occupant.direct_jid is not None
            else str(occupant.conversation_jid)
        )

    def asdict(self, return_passwords=True):
        """
        :return: The state of the client as a flat dictionary.
        """
        return {
            "room_invites": list(self.room_invites.keys()),
            "user_invites": list(self.user_invites.keys()),
            "users": list({*self.conversations.keys(), *self.users}),
            "rooms": {
                room_id: {
                    "room_id": str(room.jid),
                    "joined": room.muc_joined,
                    "state": room.muc_state.name,
                    "nick": room.me.nick if room.me else None,
                    **({"password": room.muc_password} if return_passwords else {}),
                    "members": {
                        self._occupant_user_id(member): {
                            "user_id": self._occupant_user_id(member),
                            "nick": member.nick,
                            "affiliation": member.affiliation,
                            "role": member.role,
                            "is_self": member.is_self,
                            "available": member.presence_state.available,
                            "state": (
                                "available"
                                if not member.presence_state.show.value
                                else member.presence_state.show.value
                            ),
                        }
                        for member in room.members
                    },
                }
                for room_id, room in self.rooms.items()
            },
        }

    def serialize(self) -> SerializedState:
        """
        :return: The JSON-friendly dehydrated representation of the state, which
            can be restored across restarts.
        """
        return SerializedState(**self.asdict())



from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Dict, Optional, Sequence, Union

import rtmidi

from platypush.context import get_bus
from platypush.message.event.midi import (
    MidiDeviceAddedEvent,
    MidiDeviceConnectedEvent,
    MidiDeviceDisconnectedEvent,
    MidiDeviceRemovedEvent,
)


class MidiDeviceType(Enum):
    """
    Enum for MIDI device types.
    """

    INPUT = 'input'
    OUTPUT = 'output'


@dataclass
class MidiDevice:
    """
    Data class for MIDI devices.
    """

    name: str
    port: int
    device_type: MidiDeviceType
    midi_in: Optional[rtmidi.MidiIn] = None  # type: ignore
    midi_out: Optional[rtmidi.MidiOut] = None  # type: ignore
    callback: Optional[
        Callable[
            [Optional["MidiDevice"]],
            Callable[[Sequence[int], Optional[Any]], Optional[Any]],
        ]
    ] = None

    def open(self):
        """
        Open the MIDI device.
        """
        if self.device_type == MidiDeviceType.INPUT and not self.midi_in:
            get_bus().post(MidiDeviceConnectedEvent(device=self.name, port=self.port))
            self.midi_in = rtmidi.MidiIn()  # type: ignore
            self.midi_in.open_port(self.port)
            if self.callback:
                self.midi_in.set_callback(self.callback(self))
        elif self.device_type == MidiDeviceType.OUTPUT and not self.midi_out:
            self.midi_out = rtmidi.MidiOut()  # type: ignore
            self.midi_out.open_port(self.port)

    def close(self):
        """
        Close the MIDI device.
        """
        if self.midi_in:
            self.midi_in.close_port()
            self.midi_in = None
            get_bus().post(
                MidiDeviceDisconnectedEvent(device=self.name, port=self.port)
            )

        if self.midi_out:
            self.midi_out.close_port()
            self.midi_out = None
            get_bus().post(
                MidiDeviceDisconnectedEvent(device=self.name, port=self.port)
            )

    def is_open(self) -> bool:
        """
        Check if the MIDI device is open.

        :returns: True if the MIDI device is open, False otherwise.
        """
        return (self.midi_in is not None and self.midi_in.is_port_open()) or (
            self.midi_out is not None and self.midi_out.is_port_open()
        )


class MidiDevices:
    """
    Class to manage MIDI devices.
    """

    def __init__(self):
        self._devices_by_port = {}
        self._devices_by_name = {}
        self._devices_by_type = {MidiDeviceType.INPUT: {}, MidiDeviceType.OUTPUT: {}}

    def add(self, device: MidiDevice):
        """
        Add a MIDI device to the list of available devices.

        :param device: The MIDI device to add.
        """
        if (
            device.port not in self._devices_by_port
            or self._devices_by_port[device.port].name != device.name
        ):
            self._devices_by_port[device.port] = device

        if device.name not in self._devices_by_name:
            get_bus().post(MidiDeviceAddedEvent(device=device.name, port=device.port))
            self._devices_by_name[device.name] = device

        if device.port not in self._devices_by_type[device.device_type]:
            self._devices_by_type[device.device_type][device.name] = device

    def remove(self, device: MidiDevice):
        """
        Remove a MIDI device from the list of available devices.

        :param device: The MIDI device to remove.
        """
        device.close()
        del self._devices_by_port[device.port]
        del self._devices_by_name[device.name]
        del self._devices_by_type[device.device_type][device.name]
        get_bus().post(MidiDeviceRemovedEvent(device=device.name, port=device.port))

    def by_port(self, port: int) -> Optional[MidiDevice]:
        """
        Get a MIDI device by port.

        :param port: The port of the MIDI device.
        :returns: The MIDI device if found, None otherwise.
        """
        return self._devices_by_port.get(port)

    def by_name(self, name: str) -> Optional[MidiDevice]:
        """
        Get a MIDI device by name.

        :param name: The name of the MIDI device.
        :returns: The MIDI device if found, None otherwise.
        """
        return self._devices_by_name.get(name)

    def by_type(self, device_type: MidiDeviceType) -> Dict[str, MidiDevice]:
        """
        Get all MIDI devices of a certain type.

        :param device_type: The type of the MIDI devices.
        :returns: A list of MIDI devices of the specified type.
        """
        return self._devices_by_type[device_type]

    def get(self, device: Union[str, int]) -> Optional[MidiDevice]:
        """
        Get a MIDI device by name or port.

        :param device: The name or port of the MIDI device.
        :returns: The MIDI device if found, None otherwise.
        """
        if isinstance(device, str):
            return self.by_name(device)
        if isinstance(device, int):
            return self.by_port(device)
        return None

    def close(self):
        """
        Close and clear all MIDI devices.
        """
        devices = list(self._devices_by_port.values())
        for dev in devices:
            self.remove(dev)

        self._devices_by_port.clear()
        self._devices_by_name.clear()
        self._devices_by_type = {MidiDeviceType.INPUT: {}, MidiDeviceType.OUTPUT: {}}


# vim:sw=4:ts=4:et:



from sqlalchemy import (
    Column,
    ForeignKey,
    Float,
    Integer,
)

from platypush.common.db import is_defined

from .devices import Device


if not is_defined('ping_host'):

    class PingHost(Device):
        """
        Entity that maps a generic host that can be pinged.
        """

        __tablename__ = 'ping_host'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )

        min = Column(Float)
        max = Column(Float)
        avg = Column(Float)
        mdev = Column(Float)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from .._base import BaseBluetoothClass, ClassProperty
from ._major import MajorDeviceClass


class MinorDeviceClass(BaseBluetoothClass):
    """
    Models Bluetooth minor device classes - see
    https://btprodspecificationrefs.blob.core.windows.net/assigned-numbers/Assigned%20Number%20Types/Assigned%20Numbers.pdf,
    Sections 2.8.2.1 - 2.8.2.9
    """

    # Computer classes
    COMPUTER_UNKNOWN = ClassProperty(
        'Unknown', 0xFC, 2, 0b000, MajorDeviceClass.COMPUTER
    )
    COMPUTER_DESKTOP = ClassProperty(
        'Desktop Workstation', 0xFC, 2, 0b001, MajorDeviceClass.COMPUTER
    )
    COMPUTER_SERVER = ClassProperty('Server', 0xFC, 2, 0b010, MajorDeviceClass.COMPUTER)
    COMPUTER_LAPTOP = ClassProperty('Laptop', 0xFC, 2, 0b011, MajorDeviceClass.COMPUTER)
    COMPUTER_HANDHELD_PDA = ClassProperty(
        'Handheld PDA', 0xFC, 2, 0b100, MajorDeviceClass.COMPUTER
    )
    COMPUTER_PALM_PDA = ClassProperty(
        'Palm-sized PDA', 0xFC, 2, 0b101, MajorDeviceClass.COMPUTER
    )
    COMPUTER_WEARABLE = ClassProperty(
        'Wearable Computer', 0xFC, 2, 0b110, MajorDeviceClass.COMPUTER
    )

    # Phone classes
    PHONE_UNKNOWN = ClassProperty('Unknown', 0xFC, 2, 0b000, MajorDeviceClass.PHONE)
    PHONE_CELLULAR = ClassProperty('Cellular', 0xFC, 2, 0b001, MajorDeviceClass.PHONE)
    PHONE_CORDLESS = ClassProperty('Cordless', 0xFC, 2, 0b010, MajorDeviceClass.PHONE)
    PHONE_SMARTPHONE = ClassProperty(
        'Smartphone', 0xFC, 2, 0b011, MajorDeviceClass.PHONE
    )
    PHONE_WIRED_MODEM = ClassProperty(
        'Wired Modem', 0xFC, 2, 0b100, MajorDeviceClass.PHONE
    )
    PHONE_ISDN_ACCESS = ClassProperty(
        'ISDN Access Point', 0xFC, 2, 0b101, MajorDeviceClass.PHONE
    )

    # LAN / Access Point classes
    AP_USAGE_0 = ClassProperty('Fully Available', 0xE0, 5, 0b000, MajorDeviceClass.AP)
    AP_USAGE_1_17 = ClassProperty(
        '1 - 17% Utilized', 0xE0, 5, 0b001, MajorDeviceClass.AP
    )
    AP_USAGE_17_33 = ClassProperty(
        '17 - 33% Utilized', 0xE0, 5, 0b010, MajorDeviceClass.AP
    )
    AP_USAGE_33_50 = ClassProperty(
        '33 - 50% Utilized', 0xE0, 5, 0b011, MajorDeviceClass.AP
    )
    AP_USAGE_50_67 = ClassProperty(
        '50 - 67% Utilized', 0xE0, 5, 0b100, MajorDeviceClass.AP
    )
    AP_USAGE_67_83 = ClassProperty(
        '67 - 83% Utilized', 0xE0, 5, 0b101, MajorDeviceClass.AP
    )
    AP_USAGE_83_99 = ClassProperty(
        '83 - 99% Utilized', 0xE0, 5, 0b110, MajorDeviceClass.AP
    )
    AP_USAGE_100 = ClassProperty(
        'No Service Available', 0xE0, 5, 0b111, MajorDeviceClass.AP
    )

    # Multimedia classes
    MULTIMEDIA_HEADSET = ClassProperty(
        'Headset', 0xFC, 2, 0b000001, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_HANDS_FREE = ClassProperty(
        'Hands-free Device', 0xFC, 2, 0b000010, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_MICROPHONE = ClassProperty(
        'Microphone', 0xFC, 2, 0b000100, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_LOUDSPEAKER = ClassProperty(
        'Loudspeaker', 0xFC, 2, 0b000101, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_HEADPHONES = ClassProperty(
        'Headphones', 0xFC, 2, 0b000110, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_PORTABLE_AUDIO = ClassProperty(
        'Portable Audio', 0xFC, 2, 0b000111, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_CAR_AUDIO = ClassProperty(
        'Car Audio', 0xFC, 2, 0b001000, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_SET_TOP_BOX = ClassProperty(
        'Set-top Box', 0xFC, 2, 0b001001, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_HIFI_AUDIO = ClassProperty(
        'HiFi Audio Device', 0xFC, 2, 0b001010, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_VCR = ClassProperty(
        'VCR', 0xFC, 2, 0b001011, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_VIDEO_CAMERA = ClassProperty(
        'Video Camera', 0xFC, 2, 0b001100, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_CAMCODER = ClassProperty(
        'Camcoder', 0xFC, 2, 0b001101, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_VIDEO_MONITOR = ClassProperty(
        'Video Monitor', 0xFC, 2, 0b001110, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_VIDEO_DISPLAY_AND_LOUDSPEAKER = ClassProperty(
        'Video Display and Loudspeaker', 0xFC, 2, 0b001111, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_VIDEO_CONFERENCING = ClassProperty(
        'Video Conferencing', 0xFC, 2, 0b010000, MajorDeviceClass.MULTIMEDIA
    )
    MULTIMEDIA_GAMING_TOY = ClassProperty(
        'Gaming / Toy', 0xFC, 2, 0b010010, MajorDeviceClass.MULTIMEDIA
    )

    # Peripheral classes
    PERIPHERAL_UNKNOWN = ClassProperty(
        'Unknown', 0xFC, 2, 0b000000, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_KEYBOARD = ClassProperty(
        'Keyboard', 0xC0, 6, 0b01, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_POINTER = ClassProperty(
        'Pointing Device', 0xC0, 6, 0b10, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_KEYBOARD_POINTER = ClassProperty(
        'Combo Keyboard/Pointing Device', 0xC0, 6, 0b11, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_JOYSTICK = ClassProperty(
        'Joystick', 0x3C, 2, 0b0001, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_GAMEPAD = ClassProperty(
        'Gamepad', 0x3C, 2, 0b0010, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_REMOTE_CONTROL = ClassProperty(
        'Remote Control', 0x3C, 2, 0b0011, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_SENSOR = ClassProperty(
        'Sensing Device', 0x3C, 2, 0b0100, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_DIGIT_TABLET = ClassProperty(
        'Digitizer Tablet', 0x3C, 2, 0b0101, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_CARD_READER = ClassProperty(
        'Card Reader', 0x3C, 2, 0b0110, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_DIGITAL_PEN = ClassProperty(
        'Card Reader', 0x3C, 2, 0b0111, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_SCANNER = ClassProperty(
        'Handheld Scanner', 0x3C, 2, 0b1000, MajorDeviceClass.PERIPHERAL
    )
    PERIPHERAL_GESTURES = ClassProperty(
        'Handheld Gesture Input Device', 0x3C, 2, 0b1001, MajorDeviceClass.PERIPHERAL
    )

    # Imaging classes
    IMAGING_UNKNOWN = ClassProperty(
        'Unknown', 0xFC, 2, 0b000000, MajorDeviceClass.IMAGING
    )
    IMAGING_DISPLAY = ClassProperty(
        'Display', 0xF0, 4, 0b0001, MajorDeviceClass.IMAGING
    )
    IMAGING_CAMERA = ClassProperty('Camera', 0xF0, 4, 0b0010, MajorDeviceClass.IMAGING)
    IMAGING_SCANNER = ClassProperty(
        'Scanner', 0xF0, 4, 0b0100, MajorDeviceClass.IMAGING
    )
    IMAGING_PRINTER = ClassProperty(
        'Printer', 0xF0, 4, 0b1000, MajorDeviceClass.IMAGING
    )

    # Wearable classes
    WEARABLE_UNKNOWN = ClassProperty(
        'Unknown', 0xFC, 2, 0b000000, MajorDeviceClass.WEARABLE
    )
    WEARABLE_WRISTWATCH = ClassProperty(
        'Wristwatch', 0xFC, 2, 0b000001, MajorDeviceClass.WEARABLE
    )
    WEARABLE_PAGER = ClassProperty(
        'Pager', 0xFC, 2, 0b000010, MajorDeviceClass.WEARABLE
    )
    WEARABLE_JACKET = ClassProperty(
        'Jacket', 0xFC, 2, 0b000011, MajorDeviceClass.WEARABLE
    )
    WEARABLE_HELMET = ClassProperty(
        'Helmet', 0xFC, 2, 0b000100, MajorDeviceClass.WEARABLE
    )
    WEARABLE_GLASSES = ClassProperty(
        'Glasses', 0xFC, 2, 0b000101, MajorDeviceClass.WEARABLE
    )

    # Toy classes
    TOY_UNKNOWN = ClassProperty('Unknown', 0xFC, 2, 0b000000, MajorDeviceClass.TOY)
    TOY_ROBOT = ClassProperty('Robot', 0xFC, 2, 0b000001, MajorDeviceClass.TOY)
    TOY_VEHICLE = ClassProperty('Vehicle', 0xFC, 2, 0b000010, MajorDeviceClass.TOY)
    TOY_DOLL = ClassProperty(
        'Doll / Action Figure', 0xFC, 2, 0b000011, MajorDeviceClass.TOY
    )
    TOY_CONTROLLER = ClassProperty(
        'Controller', 0xFC, 2, 0b000100, MajorDeviceClass.TOY
    )
    TOY_GAME = ClassProperty('Game', 0xFC, 2, 0b000101, MajorDeviceClass.TOY)

    # Health classes
    HEALTH_UNKNOWN = ClassProperty(
        'Unknown', 0xFC, 2, 0b000000, MajorDeviceClass.HEALTH
    )
    HEALTH_BLOOD_PRESSURE = ClassProperty(
        'Blood Pressure Monitor', 0xFC, 2, 0b000001, MajorDeviceClass.HEALTH
    )
    HEALTH_THERMOMETER = ClassProperty(
        'Thermometer', 0xFC, 2, 0b000010, MajorDeviceClass.HEALTH
    )
    HEALTH_SCALE = ClassProperty(
        'Weighing Scale', 0xFC, 2, 0b000011, MajorDeviceClass.HEALTH
    )
    HEALTH_GLUCOSE = ClassProperty(
        'Glucose Meter', 0xFC, 2, 0b000100, MajorDeviceClass.HEALTH
    )
    HEALTH_OXIMETER = ClassProperty(
        'Pulse Oximeter', 0xFC, 2, 0b000101, MajorDeviceClass.HEALTH
    )
    HEALTH_PULSE = ClassProperty(
        'Heart Rate/Pulse Monitor', 0xFC, 2, 0b000110, MajorDeviceClass.HEALTH
    )
    HEALTH_DISPLAY = ClassProperty(
        'Health Data Display', 0xFC, 2, 0b000111, MajorDeviceClass.HEALTH
    )
    HEALTH_STEPS = ClassProperty(
        'Step Counter', 0xFC, 2, 0b001000, MajorDeviceClass.HEALTH
    )
    HEALTH_COMPOSITION = ClassProperty(
        'Body Composition Analyzer', 0xFC, 2, 0b001001, MajorDeviceClass.HEALTH
    )
    HEALTH_PEAK_FLOW = ClassProperty(
        'Peak Flow Monitor', 0xFC, 2, 0b001010, MajorDeviceClass.HEALTH
    )
    HEALTH_MEDICATION = ClassProperty(
        'Medication Monitor', 0xFC, 2, 0b001011, MajorDeviceClass.HEALTH
    )
    HEALTH_KNEE_PROSTHESIS = ClassProperty(
        'Knee Prosthesis', 0xFC, 2, 0b001100, MajorDeviceClass.HEALTH
    )
    HEALTH_ANKLE_PROSTHESIS = ClassProperty(
        'Ankle Prosthesis', 0xFC, 2, 0b001101, MajorDeviceClass.HEALTH
    )
    HEALTH_GENERIC = ClassProperty(
        'Generic Health Manager', 0xFC, 2, 0b001110, MajorDeviceClass.HEALTH
    )
    HEALTH_MOBILITY = ClassProperty(
        'Personal Mobility Device', 0xFC, 2, 0b001111, MajorDeviceClass.HEALTH
    )



import logging
from typing import Dict, Iterable, Optional, Tuple

from sqlalchemy.orm import Session

from platypush.entities._base import Entity, EntityMapping
from platypush.entities._engine.repo.db import EntitiesDb
from platypush.entities._engine.repo.helpers import get_parent
from platypush.entities._engine.repo.merger import EntitiesMerger

logger = logging.getLogger('entities')


class EntitiesRepository:
    """
    This object is used to get and save entities. It wraps the database
    connection.
    """

    def __init__(self):
        self._db = EntitiesDb()
        self._merge = EntitiesMerger()

    def get(
        self, session: Session, entities: Iterable[Entity]
    ) -> Dict[Tuple[str, str], Entity]:
        """
        Given a set of entity objects, it returns those that already exist
        (or have the same ``entity_key``).
        """
        return self._db.fetch(session, entities)

    def save(self, *entities: Entity) -> Iterable[Entity]:
        """
        Perform an upsert of entities after merging duplicates and rebuilding
        the taxonomies.
        """

        with self._db.get_session(
            locked=True,
            autoflush=False,
            autocommit=False,
            expire_on_commit=False,
        ) as session:
            merged_entities = self._merge(
                session,
                entities,
                existing_entities=self._fetch_all_and_flatten(session, entities),
            )

            merged_entities = self._db.upsert(session, merged_entities)

        return merged_entities

    def _fetch_all_and_flatten(
        self,
        session: Session,
        entities: Iterable[Entity],
    ) -> EntityMapping:
        """
        Given a collection of entities, retrieves their persisted instances
        (lookup is performed by ``entity_key``), and it also recursively
        expands their relationships, so the session is updated with the latest
        persisted versions of all the objects in the hierarchy.

        :return: An ``entity_key -> entity`` mapping.
        """
        expanded_entities = {}
        for entity in entities:
            root_entity = self._get_root_entity(session, entity)
            expanded_entities.update(self._expand_children([root_entity]))
            expanded_entities.update(self._expand_children([entity]))

        return self.get(session, expanded_entities.values())

    @classmethod
    def _expand_children(
        cls,
        entities: Iterable[Entity],
        all_entities: Optional[EntityMapping] = None,
    ) -> EntityMapping:
        """
        Recursively expands and flattens all the children of a set of entities
        into an ``entity_key -> entity`` mapping.
        """
        all_entities = all_entities or {}
        for entity in entities:
            all_entities[entity.entity_key] = entity
            cls._expand_children(entity.children, all_entities)

        return all_entities

    def _get_root_entity(self, session: Session, entity: Entity) -> Entity:
        """
        Retrieve the root entity (i.e. the one with a null parent) of an
        entity.
        """
        parent = entity
        while parent:
            parent = get_parent(session, entity)
            if parent:
                entity = parent

        return entity



"""
Platyvenv is a helper script that allows you to automatically create a
virtual environment for Platypush starting from a configuration file.
"""

from contextlib import contextmanager
import logging
import os
import shutil
import subprocess
import sys
import tempfile
import textwrap
from typing import Generator, Sequence
import venv

from platypush.builder import BaseBuilder
from platypush.config import Config
from platypush.utils.manifest import (
    Dependencies,
    InstallContext,
)

logger = logging.getLogger()


class VenvBuilder(BaseBuilder):
    """
    Build a virtual environment from on a configuration file.
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs['install_context'] = InstallContext.DOCKER
        super().__init__(*args, **kwargs)

    @classmethod
    def get_name(cls):
        return "platyvenv"

    @classmethod
    def get_description(cls):
        return "Build a Platypush virtual environment from a configuration file."

    @property
    def _pip_cmd(self) -> Sequence[str]:
        """
        :return: The pip install command to use for the selected environment.
        """
        return (
            os.path.join(self.output, 'bin', 'python'),
            '-m',
            'pip',
            'install',
            '-U',
            '--no-cache-dir',
            '--no-input',
        )

    def _install_system_packages(self, deps: Dependencies):
        """
        Install the required system packages.
        """
        for cmd in deps.to_pkg_install_commands():
            logger.info('Installing system packages: %s', cmd)
            subprocess.run(cmd, shell=True, check=True)

    def _run_before_commands(self, deps: Dependencies):
        """
        Runs any commands that should be executed before the installation.
        """
        for cmd in deps.before:
            logger.info('Running: %s', cmd)
            subprocess.run(cmd, shell=True, check=True)

    def _run_after_commands(self, deps: Dependencies):
        """
        Runs any commands that should be executed after the installation.
        """
        for cmd in deps.after:
            logger.info('Running: %s', cmd)
            subprocess.run(cmd, shell=True, check=True)

    @contextmanager
    def _prepare_src_dir(self) -> Generator[str, None, None]:
        """
        Prepare the source directory used to install the virtual enviornment.

        If platyvenv is launched from a local checkout of the Platypush source
        code, then that checkout will be used.

        Otherwise, the source directory will be cloned from git into a
        temporary folder.
        """
        setup_py_path = os.path.join(os.getcwd(), 'setup.py')
        if os.path.isfile(setup_py_path):
            logger.info('Using local checkout of the Platypush source code')
            yield os.getcwd()
        else:
            checkout_dir = tempfile.mkdtemp(prefix='platypush-', suffix='.git')
            logger.info('Cloning Platypush source code from git into %s', checkout_dir)
            subprocess.call(
                [
                    'git',
                    'clone',
                    '--recursive',
                    'https://github.com/BlackLight/platypush.git',
                    checkout_dir,
                ]
            )

            pwd = os.getcwd()
            os.chdir(checkout_dir)
            subprocess.call(['git', 'checkout', self.gitref])
            yield checkout_dir

            os.chdir(pwd)
            logger.info('Cleaning up %s', checkout_dir)
            shutil.rmtree(checkout_dir, ignore_errors=True)

    def _prepare_venv(self) -> None:
        """
        Install the virtual environment under the configured output.
        """
        logger.info('Creating virtual environment under %s...', self.output)

        venv.create(
            self.output,
            symlinks=True,
            with_pip=True,
            upgrade_deps=True,
        )

        logger.info('Installing base Python dependencies under %s...', self.output)
        subprocess.call([*self._pip_cmd, 'pip', '.'])

        # Add <output>/bin to the PATH
        os.environ['PATH'] = os.path.join(self.output, 'bin') + ':' + os.environ['PATH']

    def _install_extra_pip_packages(self, deps: Dependencies):
        """
        Install the extra pip dependencies inferred from the configured
        extensions.
        """
        pip_deps = list(deps.to_pip_install_commands(full_command=False))
        if not pip_deps:
            return

        logger.info(
            'Installing extra pip dependencies under %s: %s',
            self.output,
            ' '.join(pip_deps),
        )

        subprocess.call([*self._pip_cmd, *pip_deps])

    def build(self):
        """
        Build a Dockerfile based on a configuration file.
        """
        Config.init(self.cfgfile)

        deps = Dependencies.from_config(
            self.cfgfile,
            install_context=InstallContext.VENV,
        )

        self._run_before_commands(deps)
        self._install_system_packages(deps)

        with self._prepare_src_dir():
            self._prepare_venv()

        self._install_extra_pip_packages(deps)
        self._run_after_commands(deps)

        self._print_instructions(
            textwrap.dedent(
                f"""
                Virtual environment created under {self.output}.
                To run the application:

                    source {self.output}/bin/activate
                    platypush -c {self.cfgfile} {
                        "--device_id " + self.device_id if self.device_id else ""
                    }

                Platypush requires a Redis instance. If you don't want to use a
                stand-alone server, you can pass the --start-redis option to
                the executable (optionally with --redis-port).

                Platypush will then start its own local instance and it will
                terminate it once the application is stopped.
                """
            )
        )


def main():
    """
    Generates a virtual environment based on the configuration file.
    """
    VenvBuilder.from_cmdline(sys.argv[1:]).build()
    return 0


if __name__ == '__main__':
    sys.exit(main())


# vim:sw=4:ts=4:et:



import os
from typing import Optional, Sequence

from platypush.context import get_plugin
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.assistant import AssistantPlugin
from platypush.plugins.tts.picovoice import TtsPicovoicePlugin

from ._assistant import Assistant
from ._state import AssistantState


# pylint: disable=too-many-ancestors
class AssistantPicovoicePlugin(AssistantPlugin, RunnablePlugin):
    r"""
    A voice assistant that runs on your device, based on the `Picovoice
    <https://picovoice.ai/>`_ engine.

    Picovoice is a suite of on-device voice technologies that include:

        * **Porcupine**: wake-word engine, if you want the device to listen for
          a specific wake word in order to start the assistant.

        * **Cheetah**: speech-to-text engine, if you want your voice
          interactions to be transcribed into free text - either
          programmatically or when triggered by the wake word. Or:

        * **Rhino**: intent recognition engine, if you want to extract *intents*
          out of your voice commands - for instance, the phrase "set the living
          room temperature to 20 degrees" could be mapped to the intent with the
          following parameters: ``intent``: ``set_temperature``, ``room``:
          ``living_room``, ``temperature``: ``20``.

        * **Leopard**: speech-to-text engine aimed at offline transcription of
          audio files rather than real-time transcription.

        * **Orca**: text-to-speech engine, if you want to create your custom
          logic to respond to user's voice commands and render the responses as
          audio.

    This plugin is a wrapper around the Picovoice engine that allows you to
    run your custom voice-based conversational flows on your device.

    Getting a Picovoice account and access key
    -------------------------------------------

    You can get your personal access key by signing up at the `Picovoice
    console <https://console.picovoice.ai/>`_. You may be asked to submit a
    reason for using the service (feel free to mention a personal Platypush
    integration), and you will receive your personal access key.

    If prompted to select the products you want to use, make sure to select
    the ones from the Picovoice suite that you want to use with this plugin.


    Hotword detection
    -----------------

    The hotword detection engine is based on `Porcupine
    <https://picovoice.ai/platform/porcupine/>`_.

    If enabled through the ``hotword_enabled`` parameter (default: True), the
    assistant will listen for a specific wake word before starting the
    speech-to-text or intent recognition engines. You can specify custom models
    for your hotword (e.g. on the same device you may use "Alexa" to trigger the
    speech-to-text engine in English, "Computer" to trigger the speech-to-text
    engine in Italian, and "Ok Google" to trigger the intent recognition engine.

    You can also create your custom hotword models using the `Porcupine console
    <https://console.picovoice.ai/ppn>`_.

    If ``hotword_enabled`` is set to True, you must also specify the
    ``keywords`` parameter with the list of keywords that you want to listen
    for, and optionally the ``keyword_paths`` parameter with the paths to the
    any custom hotword models that you want to use. If ``hotword_enabled`` is
    set to False, then the assistant won't start listening for speech after the
    plugin is started, and you will need to programmatically start the
    conversation by calling the :meth:`.start_conversation` action, or trigger
    it from the UI.

    When a wake-word is detected, the assistant will emit a
    :class:`platypush.message.event.assistant.HotwordDetectedEvent` event that
    you can use to build your custom logic. For example:

      .. code-block:: python

        import time

        from platypush import when, run
        from platypush.message.event.assistant import HotwordDetectedEvent

        # Turn on a light for 5 seconds when the hotword "Alexa" is detected
        @when(HotwordDetectedEvent, hotword='Alexa')
        def on_hotword_detected(event: HotwordDetectedEvent, **context):
            run("light.hue.on", lights=["Living Room"])
            time.sleep(5)
            run("light.hue.off", lights=["Living Room"])

    By default, the assistant will start listening for speech after the hotword
    if either ``stt_enabled`` or ``intent_model_path`` are set. If you don't
    want the assistant to start listening for speech after the hotword is
    detected (for example because you want to build your custom response flows,
    or trigger the speech detection using different models depending on the
    hotword that is used, or because you just want to detect hotwords but not
    speech), then you can also set the ``start_conversation_on_hotword``
    parameter to ``False``. If that is the case, then you can programmatically
    start the conversation by calling the :meth:`.start_conversation` method in
    your event hooks:

      .. code-block:: python

        from platypush import when, run
        from platypush.message.event.assistant import HotwordDetectedEvent

        # Start a conversation using the Italian language model when the
        # "Buongiorno" hotword is detected
        @when(HotwordDetectedEvent, hotword='Buongiorno')
        def on_it_hotword_detected(event: HotwordDetectedEvent, **context):
            event.assistant.start_conversation(model_file='path/to/it.pv')

    Speech-to-text
    --------------

    The speech-to-text engine is based on `Cheetah
    <https://picovoice.ai/docs/cheetah/>`_.

    If enabled through the ``stt_enabled`` parameter (default: True), the
    assistant will transcribe the voice commands into text when a conversation
    is started either programmatically through :meth:`.start_conversation` or
    when the hotword is detected.

    It will emit a
    :class:`platypush.message.event.assistant.SpeechRecognizedEvent` when some
    speech is detected, and you can hook to that event to build your custom
    logic:

      .. code-block:: python

        from platypush import when, run
        from platypush.message.event.assistant import SpeechRecognizedEvent

        # Turn on a light when the phrase "turn on the lights" is detected.
        # Note that we can leverage regex-based pattern matching to be more
        # flexible when matching the phrases. For example, the following hook
        # will be matched when the user says "turn on the lights", "turn on
        # lights", "lights on", "lights on please", "turn on light" etc.
        @when(SpeechRecognizedEvent, phrase='turn on (the)? lights?')
        def on_turn_on_lights(event: SpeechRecognizedEvent, **context):
            run("light.hue.on")

    You can also leverage context extraction through the ``${}`` syntax on the
    hook to extract specific tokens from the event that can be passed to your
    event hook. For example:

      .. code-block:: python

        from platypush import when, run
        from platypush.message.event.assistant import SpeechRecognizedEvent

        @when(SpeechRecognizedEvent, phrase='play ${title} by ${artist}')
        def on_play_track_command(
            event: SpeechRecognizedEvent, title: str, artist: str, **context
        ):
            results = run(
                "music.mopidy.search",
                filter={"title": title, "artist": artist}
            )

            if not results:
                event.assistant.render_response(f"Couldn't find {title} by {artist}")
                return

            run("music.mopidy.play", resource=results[0]["uri"])

    Speech-to-intent
    ----------------

    The intent recognition engine is based on `Rhino
    <https://picovoice.ai/docs/rhino/>`_.

    *Intents* are snippets of unstructured transcribed speech that can be
    matched to structured actions.

    Unlike with hotword and speech-to-text detection, you need to provide a
    custom model for intent detection. You can create your custom model using
    the `Rhino console <https://console.picovoice.ai/rhn>`_.

    When an intent is detected, the assistant will emit a
    :class:`platypush.message.event.assistant.IntentRecognizedEvent` that can
    be listened.

    For example, you can train a model to control groups of smart lights by
    defining the following slots on the Rhino console:

        - ``device_state``: The new state of the device (e.g. with ``on`` or
          ``off`` as supported values)

        - ``room``: The name of the room associated to the group of lights to
          be controlled (e.g. ``living room``, ``kitchen``, ``bedroom``)

    You can then define a ``lights_ctrl`` intent with the following expressions:

        - "turn ``$device_state:state`` the lights"
        - "turn ``$device_state:state`` the ``$room:room`` lights"
        - "turn the lights ``$device_state:state``"
        - "turn the ``$room:room`` lights ``$device_state:state``"
        - "turn ``$room:room`` lights ``$device_state:state``"

    This intent will match any of the following phrases:

        - "*turn on the lights*"
        - "*turn off the lights*"
        - "*turn the lights on*"
        - "*turn the lights off*"
        - "*turn on the living room lights*"
        - "*turn off the living room lights*"
        - "*turn the living room lights on*"
        - "*turn the living room lights off*"

    And it will extract any slots that are matched in the phrases in the
    :class:`platypush.message.event.assistant.IntentRecognizedEvent` event.

    Train the model, download the context file, and pass the path on the
    ``intent_model_path`` parameter.

    You can then register a hook to listen to a specific intent:

      .. code-block:: python

        from platypush import when, run
        from platypush.message.event.assistant import IntentRecognizedEvent

        @when(IntentRecognizedEvent, intent='lights_ctrl', slots={'state': 'on'})
        def on_turn_on_lights(event: IntentRecognizedEvent, **context):
            room = event.slots.get('room')
            if room:
                run("light.hue.on", groups=[room])
            else:
                run("light.hue.on")

    Note that if both ``stt_enabled`` and ``intent_model_path`` are set, then
    both the speech-to-text and intent recognition engines will run in parallel
    when a conversation is started.

    The intent engine is usually faster, as it has a smaller set of intents to
    match and doesn't have to run a full speech-to-text transcription. This means that,
    if an utterance matches both a speech-to-text phrase and an intent, the
    :class:`platypush.message.event.assistant.IntentRecognizedEvent` event is emitted
    (and not :class:`platypush.message.event.assistant.SpeechRecognizedEvent`).

    This may not be always the case though. So it may be a good practice to
    also provide a fallback
    :class:`platypush.message.event.assistant.SpeechRecognizedEvent` hook to
    catch the text if the speech is not recognized as an intent:

      .. code-block:: python

        from platypush import when, run
        from platypush.message.event.assistant import SpeechRecognizedEvent

        @when(SpeechRecognizedEvent, phrase='turn ${state} (the)? ${room} lights?')
        def on_turn_on_lights(event: SpeechRecognizedEvent, phrase, room, **context):
            if room:
                run("light.hue.on", groups=[room])
            else:
                run("light.hue.on")

    Text-to-speech
    --------------

    The text-to-speech engine is based on `Orca
    <https://picovoice.ai/docs/orca/>`_.

    It is not directly implemented by this plugin, but the implementation is
    provided in the :class:`platypush.plugins.tts.picovoice.TtsPicovoicePlugin`
    plugin.

    You can however leverage the :meth:`.render_response` action to render some
    text as speech in response to a user command, and that in turn will leverage
    the PicoVoice TTS plugin to render the response.

    For example, the following snippet provides a hook that:

        - Listens for
          :class:`platypush.message.event.assistant.SpeechRecognizedEvent`.

        - Matches the phrase against a list of predefined commands that
          shouldn't require a response.

        - Has a fallback logic that leverages the
          :class:`platypush.plugins.openai.OpenaiPlugin` to generate a response
          for the given text and renders it as speech.

        - Has a logic for follow-on turns if the response from ChatGPT is a question.

      .. code-block:: python

        import re
        from collections import defaultdict
        from datetime import datetime as dt, timedelta
        from dateutil.parser import isoparse
        from logging import getLogger

        from platypush import hook, run
        from platypush.message.event.assistant import (
            SpeechRecognizedEvent,
            ResponseEndEvent,
        )

        logger = getLogger(__name__)

        def play_music(*_, **__):
            run("music.mopidy.play")

        def stop_music(*_, **__):
            run("music.mopidy.stop")

        def ai_assist(event: SpeechRecognizedEvent, **__):
            response = run("openai.get_response", prompt=event.phrase)
            if not response:
                return

            run("assistant.picovoice.render_response", text=response)

        # List of commands to match, as pairs of regex patterns and the
        # corresponding actions
        hooks = (
            (re.compile(r"play (the)?music", re.IGNORECASE), play_music),
            (re.compile(r"stop (the)?music", re.IGNORECASE), stop_music),
            # Fallback to the AI assistant
            (re.compile(r".*"), ai_assist),
        )

        @when(SpeechRecognizedEvent)
        def on_speech_recognized(event, **kwargs):
            for pattern, command in hooks:
                if pattern.search(event.phrase):
                    logger.info("Running voice command %s", command.__name__)
                    command(event, **kwargs)
                    break

        @when(ResponseEndEvent)
        def on_response_end(event: ResponseEndEvent, **__):
            # Check if the response is a question and start a follow-on turn if so.
            # Note that the ``openai`` plugin by default is configured to keep
            # the past interaction in a context window of ~10 minutes, so you
            # can follow up like in a real conversation.
            if event.assistant and event.response_text and event.response_text.endswith("?"):
                event.assistant.start_conversation()

    """

    def __init__(
        self,
        access_key: str,
        hotword_enabled: bool = True,
        stt_enabled: bool = True,
        keywords: Optional[Sequence[str]] = None,
        keyword_paths: Optional[Sequence[str]] = None,
        keyword_model_path: Optional[str] = None,
        speech_model_path: Optional[str] = None,
        intent_model_path: Optional[str] = None,
        endpoint_duration: Optional[float] = 0.5,
        enable_automatic_punctuation: bool = False,
        start_conversation_on_hotword: bool = True,
        audio_queue_size: int = 100,
        conversation_timeout: Optional[float] = 7.5,
        muted: bool = False,
        **kwargs,
    ):
        """
        :param access_key: Your Picovoice access key. You can get it by signing
            up at the `Picovoice console <https://console.picovoice.ai/>`.
        :param hotword_enabled: Enable the wake-word engine (default: True).
            **Note**: The wake-word engine requires you to add Porcupine to the
            products available in your Picovoice account.
        :param stt_enabled: Enable the speech-to-text engine (default: True).
            **Note**: The speech-to-text engine requires you to add Cheetah to
            the products available in your Picovoice account.
        :param keywords: List of keywords to listen for (e.g. ``alexa``, ``ok
            google``...). This is required if the wake-word engine is enabled.
            See the `Porcupine keywords repository
            <https://github.com/Picovoice/porcupine/tree/master/resources/keyword_files>`_).
            for a list of the stock keywords available. If you have a custom
            model, you can pass its path to the ``keyword_paths`` parameter and
            its filename (without the path and the platform extension) here.
        :param keyword_paths: List of paths to the keyword files to listen for.
            Custom keyword files can be created using the `Porcupine console
            <https://console.picovoice.ai/ppn>`_ and downloaded from the
            console itself.
        :param keyword_model_path: If you are using a keyword file in a
            non-English language, you can provide the path to the model file
            for its language. Model files are available for all the supported
            languages through the `Porcupine lib repository
            <https://github.com/Picovoice/porcupine/tree/master/lib/common>`_.
        :param speech_model_path: Path to the speech model file. If you are
            using a language other than English, you can provide the path to the
            model file for that language. Model files are available for all the
            supported languages through the `Cheetah repository
            <https://github.com/Picovoice/cheetah/tree/master/lib/common>`_.
            You can also use the `Speech console
            <https://console.picovoice.ai/cat>`_
            to train your custom models. You can use a base model and fine-tune
            it by boosting the detection of your own words and phrases and edit
            the phonetic representation of the words you want to detect.
        :param intent_model_path: Path to the Rhino context model. This is
            required if you want to use the intent recognition engine through
            Rhino. The context model is a file that contains a list of intents
            that can be recognized by the engine. An intent is an action or a
            class of actions that the assistant can recognize, and it can
            contain an optional number of slots to model context variables -
            e.g. temperature, lights group, location, device state etc.
            You can create your own context model using the `Rhino console
            <https://console.picovoice.ai/rhn>`_. For example, you can define a
            context file to control smart home devices by defining the
            following slots:

                - ``device_type``: The device to control (e.g. lights, music)
                - ``device_state``: The target state of the device (e.g. on,
                  off)
                - ``location``: The location of the device (e.g. living
                  room, kitchen, bedroom)
                - ``media_type``: The type of media to play (e.g. music, video)
                - ``media_state``: The state of the media (e.g. play, pause,
                  stop)

            You can then define the following intents:

                - ``device_ctrl``: Control a device state. Supported phrases:
                    - "turn ``$device_state:state`` the ``$location:location``
                      ``$device_type:device``"
                    - "turn ``$device_state:state`` the ``$device_type:device``"

                - ``media_ctrl``: Control media state. Supported phrases:
                    - "``$media_state:state`` the ``$media_type:media``"
                    - "``$media_state:state`` the ``$media_type:media`` in the
                      ``$location:location``"

            Then a phrase like "turn on the lights in the living room" would
            trigger a
            :class:`platypush.message.event.assistant.IntentRecognizedEvent` with:

                .. code-block:: json

                  {
                    "intent": "device_ctrl",
                    "slots": {
                      "type": "lights",
                      "state": "on",
                      "location": "living room"
                    }
                  }

            **Note**: The intent recognition engine requires you to add Rhino
            to the products available in your Picovoice account.
        :param endpoint_duration: If set, the assistant will stop listening when
            no speech is detected for the specified duration (in seconds) after
            the end of an utterance.
        :param enable_automatic_punctuation: Enable automatic punctuation
            insertion.
        :param start_conversation_on_hotword: If set to True (default), a speech
            detection session will be started when the hotword is detected. If
            set to False, you may want to start the conversation programmatically
            by calling the :meth:`.start_conversation` method instead, or run any
            custom logic hotword detection logic. This can be particularly useful
            when you want to run the assistant in a push-to-talk mode, or when you
            want different hotwords to trigger conversations with different models
            or languages.
        :param audio_queue_size: Maximum number of audio frames to hold in the
            processing queue. You may want to increase this value if you are
            running this integration on a slow device and/or the logs report
            audio frame drops too often. Keep in mind that increasing this value
            will increase the memory usage of the integration. Also, a higher
            value may result in higher accuracy at the cost of higher latency.
        :param conversation_timeout: Maximum time to wait for some speech to be
            detected after the hotword is detected. If no speech is detected
            within this time, the conversation will time out and the plugin will
            go back into hotword detection mode, if the mode is enabled. Default:
            7.5 seconds.
        :param muted: Set to True to start the assistant in a muted state. You will
            need to call the :meth:`.unmute` method to start the assistant listening
            for commands, or programmatically call the :meth:`.start_conversation`
            to start a conversation.
        """
        super().__init__(**kwargs)
        self._assistant = None
        self._assistant_args = {
            'stop_event': self._should_stop,
            'access_key': access_key,
            'hotword_enabled': hotword_enabled,
            'stt_enabled': stt_enabled,
            'keywords': keywords,
            'keyword_paths': (
                os.path.expanduser(keyword_path)
                for keyword_path in (keyword_paths or [])
            ),
            'keyword_model_path': (
                os.path.expanduser(keyword_model_path) if keyword_model_path else None
            ),
            'speech_model_path': (
                os.path.expanduser(speech_model_path) if speech_model_path else None
            ),
            'intent_model_path': (
                os.path.expanduser(intent_model_path) if intent_model_path else None
            ),
            'endpoint_duration': endpoint_duration,
            'enable_automatic_punctuation': enable_automatic_punctuation,
            'start_conversation_on_hotword': (
                start_conversation_on_hotword
                if (intent_model_path or stt_enabled)
                else False
            ),
            'audio_queue_size': audio_queue_size,
            'conversation_timeout': conversation_timeout,
            'muted': muted,
            'on_conversation_start': self._on_conversation_start,
            'on_conversation_end': self._on_conversation_end,
            'on_conversation_timeout': self._on_conversation_timeout,
            'on_speech_recognized': self._on_speech_recognized,
            'on_intent_matched': self._on_intent_matched,
            'on_hotword_detected': self._on_hotword_detected,
        }

    @property
    def tts(self) -> TtsPicovoicePlugin:
        p = get_plugin('tts.picovoice')
        assert p, 'Picovoice TTS plugin not configured/found'
        return p

    def _get_tts_plugin(self) -> TtsPicovoicePlugin:
        return self.tts

    def _on_response_render_start(self, text: Optional[str], *_, **__):
        if self._assistant:
            self._assistant.set_responding(True)
        return super()._on_response_render_start(text)

    def _on_response_render_end(self, *_, **__):
        if self._assistant:
            self._assistant.set_responding(False)

        return super()._on_response_render_end()

    @action
    def start_conversation(self, *_, model_file: Optional[str] = None, **__):
        """
        Programmatically start a conversation with the assistant.

        :param model_file: Override the model file to be used to detect speech
            in this conversation. If not set, the configured
            ``speech_model_path`` will be used.
        """
        if not self._assistant:
            self.logger.warning('Assistant not initialized')
            return

        if not model_file:
            model_file = self._assistant_args['speech_model_path']
        if model_file:
            model_file = os.path.expanduser(model_file)

        self._assistant.override_speech_model(model_file)
        self._assistant.state = AssistantState.DETECTING_SPEECH

    def _stop_conversation(self, *_, **__):
        super()._stop_conversation()
        if not self._assistant:
            self.logger.warning('Assistant not initialized')
            return

        self._assistant.override_speech_model(None)

        if self._assistant.hotword_enabled:
            self._assistant.state = AssistantState.DETECTING_HOTWORD
        else:
            self._assistant.state = AssistantState.IDLE

    @action
    def say(self, text: str, *args, **kwargs):
        """
        Proxy to
        :class:`platypush.plugins.tts.picovoice.TtsPicovoicePlugin.say` to
        render some text as speech through the Picovoice TTS engine.

        Extra arguments to
        :class:`platypush.plugins.tts.picovoice.TtsPicovoicePlugin.say` can be
        passed over ``args`` and ``kwargs``.

        :param text: Text to be rendered as speech.
        """
        return self.tts.say(text, *args, **kwargs)

    @action
    def transcribe(self, audio_file: str, *_, model_file: Optional[str] = None, **__):
        """
        Transcribe an audio file to text using the `Leopard
        <https://picovoice.ai/docs/leopard/>`_ engine.

        :param text: Text to be transcribed.
        :param model_file: Override the model file to be used to detect speech
            in this conversation. If not set, the configured
            ``speech_model_path`` will be used.
        :return: dict

          .. code-block:: json

            {
              "transcription": "This is a test",
              "words": [
                {
                  "word": "this",
                  "start": 0.06400000303983688,
                  "end": 0.19200000166893005,
                  "confidence": 0.9626294374465942
                },
                {
                  "word": "is",
                  "start": 0.2879999876022339,
                  "end": 0.35199999809265137,
                  "confidence": 0.9781675934791565
                },
                {
                  "word": "a",
                  "start": 0.41600000858306885,
                  "end": 0.41600000858306885,
                  "confidence": 0.9764975309371948
                },
                {
                  "word": "test",
                  "start": 0.5120000243186951,
                  "end": 0.8320000171661377,
                  "confidence": 0.9511580467224121
                }
              ]
            }

        """
        import pvleopard

        audio_file = os.path.expanduser(audio_file)
        if not model_file:
            model_file = self._assistant_args['speech_model_path']
        if model_file:
            model_file = os.path.expanduser(model_file)

        leopard = pvleopard.create(
            access_key=self._assistant_args['access_key'], model_path=model_file
        )

        transcript, words = leopard.process_file(audio_file)

        try:
            return {
                'transcription': transcript,
                'words': [
                    {
                        'word': word.word,
                        'start': word.start_sec,
                        'end': word.end_sec,
                        'confidence': word.confidence,
                    }
                    for word in words
                ],
            }
        finally:
            leopard.delete()

    @action
    def mute(self, *_, **__):
        """
        Mute the microphone. Alias for :meth:`.set_mic_mute` with
        ``muted=True``.
        """
        return self.set_mic_mute(muted=True)

    @action
    def unmute(self, *_, **__):
        """
        Unmute the microphone. Alias for :meth:`.set_mic_mute` with
        ``muted=False``.
        """
        return self.set_mic_mute(muted=False)

    @action
    def set_mic_mute(self, muted: bool):
        """
        Programmatically mute/unmute the microphone.

        :param muted: Set to True or False.
        """
        if self._assistant:
            self._assistant.set_mic_mute(muted)

        super()._on_mute_changed(muted)

    @action
    def toggle_mute(self, *_, **__):
        """
        Toggle the mic mute state.
        """
        return self.set_mic_mute(not self._is_muted)

    @action
    def send_text_query(self, *_, query: str, **__):
        """
        Send a text query to the assistant.

        This is equivalent to saying something to the assistant.

        :param query: Query to be sent.
        """
        self._on_speech_recognized(query)

    def main(self):
        while not self.should_stop():
            self.logger.info('Starting Picovoice assistant')
            with Assistant(**self._assistant_args) as self._assistant:
                try:
                    for event in self._assistant:
                        if event is not None:
                            self.logger.debug('Dequeued assistant event: %s', event)
                except KeyboardInterrupt:
                    break
                except Exception as e:
                    self.logger.error('Picovoice assistant error: %s', e, exc_info=True)
                    self.wait_stop(5)

    def stop(self):
        try:
            self.stop_conversation()
        except RuntimeError:
            pass

        super().stop()


# vim:sw=4:ts=4:et:



from enum import Enum

from marshmallow import pre_load

from .._base import SystemBaseSchema


class BatteryBaseSchema(SystemBaseSchema):
    """
    Base schema for system battery sensors.
    """

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        percent = data.pop('percent', data.pop('value', None))
        seconds_left = data.pop('secsleft', data.pop('seconds_left', None))
        data['value'] = percent / 100 if percent is not None else None
        data['seconds_left'] = None if isinstance(seconds_left, Enum) else seconds_left
        return data



import time
from typing import Optional, Dict, Iterable

from platypush.plugins import Plugin, action


class PwmPca9685Plugin(Plugin):
    """
    This plugin interacts with an `Adafruit PCA9685 <https://learn.adafruit.com/16-channel-pwm-servo-driver?view=all>`_
    circuit, a 16-channel PWM servo driver. You can leverage this plugin to control servos and motors through Platypush.

    Note that the driver for PCA9685 has been written for CircuitPython, and in order to use it on a Raspberry Pi or
    similar devices you need to
    `follow these steps <https://learn.adafruit.com/16-channel-pwm-servo-driver?view=all#python-circuitpython>`_::

        # pip3 install --upgrade adafruit-python-shell
        $ wget https://raw.githubusercontent.com/adafruit/Raspberry-Pi-Installer-Scripts/master/raspi-blinka.py
        # python3 raspi-blinka.py

    Then reboot and check that the following virtual devices are available::

        - /dev/i2c-1
        - /dev/spidev0.0
        - /dev/spidev0.1

    Finally, install the PCA9685 drivers::

        # pip3 install --upgrade adafruit-circuitpython-pca9685

    This plugin works with a PCA9685 circuit connected to the Platypush host over I2C interface.
    """

    def __init__(
        self,
        frequency: float,
        min_duty_cycle: int = 0,
        max_duty_cycle: int = 0xFFFF,
        channels: Optional[Iterable[int]] = None,
        **kwargs
    ):
        """
        :param frequency: Default PWM frequency to use for the driver, in Hz.
        :param min_duty_cycle: Minimum PWM duty cycle (you can often find it in the documentation of your device).
            Default: 0.
        :param max_duty_cycle: Maximum PWM duty cycle (you can often find it in the documentation of your device).
            Default: 0xffff.
        :param channels: Indices of the default channels to be controlled (default: all channels,
            i.e. ``[0-15]``).
        """
        super().__init__(**kwargs)
        self.frequency = frequency
        self.min_duty_cycle = min_duty_cycle
        self.max_duty_cycle = max_duty_cycle
        self.channels = channels or tuple(range(16))
        self._pca = None

    @action
    def write(
        self,
        value: Optional[int] = None,
        channels: Optional[Dict[int, float]] = None,
        frequency: Optional[float] = None,
        step: Optional[int] = None,
        step_duration: Optional[float] = None,
    ):
        """
        Send PWM values to the channels.

        :param value: Send the value to all the channels (or to all the configured default channels).
            ``value`` and ``channels`` are mutually exclusive.
        :param channels: Map of the values to be written, as a ``channel_index -> value``, where value is a real number.
            ``value`` and ``channels`` are mutually exclusive.
        :param frequency: Override default frequency.
        :param step: If set, then the PWM duty cycle will be increased/decreased by
            this much per cycle (i.e. 1/frequency). This is useful when dealing with PWM
            devices that require smooth transitions, arming sequences or ramping signals.
            If None (default) then the new PWM values will be directly written with no
            ramping logic.
        :param step_duration: If step is configured, this parameter identifies how long
            each step should last (default: ``1/frequency``).
        """
        import busio
        from board import SCL, SDA
        from adafruit_pca9685 import PCA9685

        if value is not None:
            assert self.channels, 'No default channels configured'
            channels = {i: value for i in self.channels}

        assert channels, 'Both value and channels are missing'

        i2c_bus = busio.I2C(SCL, SDA)
        pca = self._pca = self._pca or PCA9685(i2c_bus)
        pca.frequency = frequency or self.frequency
        step_duration = step_duration or 1 / pca.frequency

        if not step:
            for i, val in channels.items():
                pca.channels[i].duty_cycle = val
            return

        done = False
        cur_values = {i: channel.duty_cycle for i, channel in enumerate(pca.channels)}

        while not done:
            done = True

            for i, val in channels.items():
                if val == cur_values[i]:
                    continue

                done = False
                val = (
                    min(cur_values[i] + step, val, self.max_duty_cycle)
                    if val > pca.channels[i].duty_cycle
                    else max(cur_values[i] - step, val, self.min_duty_cycle)
                )

                pca.channels[i].duty_cycle = cur_values[i] = val

            time.sleep(step_duration)

    @action
    def get_channels(self) -> Dict[int, float]:
        """
        Get the current duty cycle value of the channels.

        :return: A map in the format ``channel_index -> value``, where ``value`` is the duty cycle of the associated
            channel, as a percentage value between 0 and 1.
        """
        if not self._pca:
            return {i: 0 for i in self.channels}

        return {i: channel.duty_cycle for i, channel in enumerate(self._pca.channels)}

    @action
    def deinit(self):
        """
        De-initialize the PCA9685 if it was previously initialized.
        """
        if not self._pca:
            self.logger.warning('The PCA9685 driver is not initialized')
            return

        self._pca.deinit()
        self._pca = None

    @action
    def reset(self):
        """
        Reset the PCA9685 (set all the channels to zero).
        """
        if not self._pca:
            self.logger.warning('The PCA9685 driver is not initialized')
            return

        self._pca.reset()



from abc import ABC
from typing import Optional, Type, Union

import aioxmpp
import aioxmpp.im.p2p

from platypush.context import get_bus
from platypush.message.event.xmpp import XmppEvent

from ._base import XmppBaseMixin


# pylint: disable=too-few-public-methods
class XmppEventMixin(XmppBaseMixin, ABC):
    """
    This mixin provides utility methods to post XMPP events.
    """

    def _post_event(self, event_type: Type[XmppEvent], *args, **kwargs):
        get_bus().post(
            event_type(
                *args,
                client_jabber_id=self._jid_to_str(
                    self._client.local_jid if self._client else self._jid
                ),
                **kwargs,
            )
        )

    def _post_user_event(
        self,
        event_type: Type[XmppEvent],
        user_id: Union[str, aioxmpp.JID],
        *args,
        **kwargs,
    ):
        if isinstance(user_id, str):
            kwargs['user_id'] = user_id
            kwargs['jid'] = user_id
        else:
            kwargs['user_id'] = self._jid_to_str(user_id)
            kwargs['jid'] = str(user_id)

        self._post_event(event_type, *args, **kwargs)

    def _post_room_event(
        self, event_type: Type[XmppEvent], room: aioxmpp.muc.Room, *args, **kwargs
    ):
        self._post_event(
            event_type, *args, room_id=self._jid_to_str(room.jid), **kwargs
        )

    def _post_user_room_event(
        self,
        event_type: Type[XmppEvent],
        room: aioxmpp.muc.Room,
        user_id: Union[str, aioxmpp.JID],
        *args,
        **kwargs,
    ):
        self._post_user_event(
            event_type,
            *args,
            user_id=user_id,
            room_id=self._jid_to_str(room.jid),
            **kwargs,
        )

    def _post_conversation_event(
        self,
        event_type: Type[XmppEvent],
        conversation: aioxmpp.im.p2p.Conversation,
        *args,
        **kwargs,
    ):
        self._post_event(
            event_type, *args, conversation_id=str(conversation.jid), **kwargs
        )

    def _post_room_occupant_event(
        self,
        event_type: Type[XmppEvent],
        room: aioxmpp.muc.Room,
        occupant: aioxmpp.muc.service.Occupant,
        *args,
        user_id: Optional[str] = None,
        **kwargs,
    ):
        self._post_user_room_event(
            event_type,
            *args,
            room=room,
            user_id=user_id
            or (
                occupant.direct_jid
                if occupant.direct_jid
                else occupant.conversation_jid
            ),
            is_self=occupant.is_self,
            **kwargs,
        )

    def _post_conversation_member_event(
        self,
        event_type: Type[XmppEvent],
        conversation: aioxmpp.im.p2p.Conversation,
        member: aioxmpp.im.p2p.Member,
        *args,
        user_id: Optional[str] = None,
        **kwargs,
    ):
        self._post_conversation_event(
            event_type,
            *args,
            conversation=conversation,
            user_id=user_id
            or (
                self._jid_to_str(member.direct_jid)
                if member.direct_jid
                else str(member.conversation_jid)
            ),
            is_self=member.is_self,
            **kwargs,
        )



from datetime import datetime
from typing import List, Optional

from platypush.message.event import Event


class NewWeatherConditionEvent(Event):
    """
    Event triggered when the weather condition changes
    """

    def __init__(
        self,
        *args,
        plugin_name: str,
        summary: Optional[str] = None,
        icon: Optional[str] = None,
        image: Optional[str] = None,
        precip_intensity: Optional[float] = None,
        precip_type: Optional[str] = None,
        temperature: Optional[float] = None,
        apparent_temperature: Optional[float] = None,
        humidity: Optional[float] = None,
        pressure: Optional[float] = None,
        rain_chance: Optional[float] = None,
        wind_speed: Optional[float] = None,
        wind_gust: Optional[float] = None,
        wind_direction: Optional[float] = None,
        cloud_cover: Optional[float] = None,
        visibility: Optional[float] = None,
        sunrise: Optional[datetime] = None,
        sunset: Optional[datetime] = None,
        units: str = 'metric',
        **kwargs,
    ):
        """
        :param plugin_name: Plugin that triggered the event.
        :param summary: Summary of the weather condition.
        :param icon: Icon representing the weather condition.
        :param image: Image URL representing the weather condition.
        :param precip_intensity: Intensity of the precipitation.
        :param precip_type: Type of precipitation.
        :param temperature: Temperature, in the configured unit system.
        :param apparent_temperature: Apparent temperature, in the configured
            unit system.
        :param humidity: Humidity percentage, between 0 and 100.
        :param pressure: Pressure, in the configured unit system.
        :param rain_chance: Chance of rain, between 0 and 100.
        :param wind_speed: Wind speed, in the configured unit system.
        :param wind_gust: Wind gust, in the configured unit system.
        :param wind_direction: Wind direction, in degrees.
        :param cloud_cover: Cloud cover percentage, between 0 and 100.
        :param visibility: Visibility, in meters.
        :param sunrise: Sunrise time.
        :param sunset: Sunset time.
        :param units: Unit system (default: metric).
        """
        super().__init__(
            *args,
            plugin_name=plugin_name,
            summary=summary,
            icon=icon,
            image=image,
            precip_intensity=precip_intensity,
            precip_type=precip_type,
            temperature=temperature,
            apparent_temperature=apparent_temperature,
            humidity=humidity,
            pressure=pressure,
            rain_chance=rain_chance,
            wind_speed=wind_speed,
            wind_gust=wind_gust,
            wind_direction=wind_direction,
            cloud_cover=cloud_cover,
            visibility=visibility,
            sunrise=sunrise,
            sunset=sunset,
            units=units,
            **kwargs,
        )


class NewPrecipitationForecastEvent(Event):
    """
    Event triggered when the precipitation forecast changes
    """

    def __init__(
        self,
        *args,
        plugin_name: Optional[str] = None,
        average: float,
        total: float,
        time_frame: int,
        **kwargs,
    ):
        super().__init__(
            *args,
            plugin_name=plugin_name,
            average=average,
            total=total,
            time_frame=time_frame,
            **kwargs,
        )


class NewWeatherForecastEvent(Event):
    """
    Event triggered when a new weather forecast is received.
    """

    def __init__(
        self,
        *args,
        plugin_name: str,
        forecast: List[dict],
        **kwargs,
    ):
        """
        :param forecast: List of weather forecast items. Format:

            .. schema:: weather.openweathermap.WeatherSchema(many=True)

        """
        super().__init__(
            *args,
            plugin_name=plugin_name,
            forecast=forecast,
            **kwargs,
        )


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from functools import partial
from typing import Any, List, Optional

from marshmallow import pre_load
from marshmallow.fields import Function
from marshmallow.validate import Range
from marshmallow_dataclass import class_schema

from platypush.schemas import EnumField
from platypush.schemas.dataclasses import DataClassSchema


class LinodeInstanceStatus(Enum):
    """
    Maps the possible states of an instance.
    """

    RUNNING = 'running'
    OFFLINE = 'offline'
    BOOTING = 'booting'
    REBOOTING = 'rebooting'
    SHUTTING_DOWN = 'shutting_down'
    PROVISIONING = 'provisioning'
    DELETING = 'deleting'
    MIGRATING = 'migrating'
    REBUILDING = 'rebuilding'
    CLONING = 'cloning'
    RESTORING = 'restoring'
    STOPPED = 'stopped'


class LinodeInstanceBackupScheduleDay(Enum):
    """
    Allowed values for ``backups.schedule.day``.
    """

    SCHEDULING = 'Scheduling'
    SUNDAY = 'Sunday'
    MONDAY = 'Monday'
    TUESDAY = 'Tuesday'
    WEDNESDAY = 'Wednesday'
    THURSDAY = 'Thursday'
    FRIDAY = 'Friday'
    SATURDAY = 'Saturday'


class LinodeInstanceBackupScheduleWindow(Enum):
    """
    Allowed values for ``backups.schedule.window``.

    The window in which your backups will be taken, in UTC. A backups window is
    a two-hour span of time in which the backup may occur.

    For example, W10 indicates that your backups should be taken between 10:00
    and 12:00.
    """

    SCHEDULING = 'Scheduling'
    W0 = 'W0'
    W2 = 'W2'
    W4 = 'W4'
    W6 = 'W6'
    W8 = 'W8'
    W10 = 'W10'
    W12 = 'W12'
    W14 = 'W14'
    W16 = 'W16'
    W18 = 'W18'
    W20 = 'W20'
    W22 = 'W22'


class FieldWithId(Function):
    """
    Field that handles values that are objects with an ``id`` attribute.
    """

    def _deserialize(self, value: Any, *_, **__) -> Optional[Any]:
        return value.id if value is not None else None

    def _serialize(self, value: Any, *_, **__) -> Optional[Any]:
        return value


class LinodeBaseSchema(DataClassSchema):
    """
    Base schema for all Linode objects.
    """

    TYPE_MAPPING = {
        LinodeInstanceStatus: partial(  # type: ignore
            EnumField, type=LinodeInstanceStatus
        ),
        LinodeInstanceBackupScheduleDay: partial(  # type: ignore
            EnumField, type=LinodeInstanceBackupScheduleDay
        ),
        LinodeInstanceBackupScheduleWindow: partial(  # type: ignore
            EnumField, type=LinodeInstanceBackupScheduleWindow
        ),
        **DataClassSchema.TYPE_MAPPING,
    }

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        from linode_api4.objects.base import MappedObject

        # Expand MappedObjects to dictionaries
        for key, value in data.items():
            if isinstance(value, MappedObject):
                data[key] = value.dict

        # NOTE Workaround for type -> instance_type not being correctly mapped
        if 'type' in data:
            data['instance_type'] = data.pop('type')

        return data


@dataclass
class LinodeInstanceSpecs:
    """
    Class that models the specifications of a Linode instance.
    """

    disk: int = field(
        metadata={
            'metadata': {
                'description': 'Allocated disk size, in MB',
                'example': 100000,
            }
        }
    )

    memory: int = field(
        metadata={
            'metadata': {
                'description': 'Allocated RAM size, in MB',
                'example': 8192,
            }
        }
    )

    cpus: int = field(
        metadata={
            'data_key': 'vcpus',
            'metadata': {
                'description': 'Number of virtual CPUs allocated to the instance',
                'example': 4,
            },
        }
    )

    gpus: int = field(
        metadata={
            'metadata': {
                'description': 'Number of GPUs allocated to the instance',
                'example': 1,
            }
        }
    )

    transfer: int = field(
        metadata={
            'metadata': {
                'description': (
                    'Number of network transfers this instance is allotted each month',
                ),
                'example': 5000,
            }
        }
    )


@dataclass
class LinodeInstanceAlerts:
    """
    Class that models the alerts configuration of a Linode instance.
    """

    cpu: int = field(
        metadata={
            'metadata': {
                'validate': Range(min=0, max=100),
                'description': (
                    'The percentage of CPU average usage over the past two hours '
                    'required to trigger an alert',
                ),
                'example': 90,
            }
        }
    )

    io: int = field(
        metadata={
            'metadata': {
                'description': (
                    'The amount of disk I/O operations per second required to '
                    'trigger an alert'
                ),
                'example': 5000,
            }
        }
    )

    network_in: int = field(
        metadata={
            'metadata': {
                'description': (
                    'The amount of incoming network traffic, in Mbit/s, '
                    'required to trigger an alert'
                ),
                'example': 10,
            }
        }
    )

    network_out: int = field(
        metadata={
            'metadata': {
                'description': (
                    'The amount of outgoing network traffic, in Mbit/s, '
                    'required to trigger an alert'
                ),
                'example': 10,
            }
        }
    )

    transfer_quota: int = field(
        metadata={
            'metadata': {
                'validate': Range(min=0, max=100),
                'description': (
                    'The percentage of network transfer that may be used before '
                    'an alert is triggered',
                ),
                'example': 80,
            }
        }
    )


@dataclass
class LinodeInstanceBackupSchedule:
    """
    Class that models the backup schedule of a Linode instance.
    """

    day: Optional[LinodeInstanceBackupScheduleDay]
    window: Optional[LinodeInstanceBackupScheduleWindow]


@dataclass
class LinodeInstanceBackups:
    """
    Class that models the backup status of a Linode instance.
    """

    available: bool
    enabled: bool = field(
        metadata={
            'metadata': {
                'description': 'Whether the backups are enabled on this instance',
                'example': True,
            }
        }
    )

    schedule: LinodeInstanceBackupSchedule
    last_successful: Optional[datetime] = field(
        metadata={
            'metadata': {
                'description': 'When the last backup was successful',
                'example': '2020-01-01T00:00:00Z',
            }
        }
    )


@dataclass
class LinodeInstance:
    """
    Class that models a Linode instance.
    """

    id: int = field(
        metadata={
            'required': True,
            'metadata': {
                'description': 'Instance ID',
                'example': 12345,
            },
        }
    )

    name: str = field(
        metadata={
            'required': True,
            'data_key': 'label',
            'metadata': {
                'description': 'Instance name',
                'example': 'my-instance',
            },
        },
    )

    instance_type: str = field(
        metadata={
            'marshmallow_field': FieldWithId(),
            'metadata': {
                'description': 'Instance type',
                'example': 'g6-standard-4',
            },
        }
    )

    ipv4_addresses: List[str] = field(
        metadata={
            'data_key': 'ipv4',
            'metadata': {
                'description': 'List of IPv4 addresses associated with this instance',
                'example': '["1.2.3.4"]',
            },
        }
    )

    ipv6_address: str = field(
        metadata={
            'data_key': 'ipv6',
            'metadata': {
                'description': 'IPv6 address associated with this instance',
                'example': '1234:5678::9abc:def0:1234:5678/128',
            },
        }
    )

    group: str = field(
        metadata={
            'metadata': {
                'description': 'Group the instance belongs to',
                'example': 'my-group',
            }
        }
    )

    status: LinodeInstanceStatus = field(
        metadata={
            'metadata': {
                'description': 'Instance status',
                'example': 'running',
            }
        }
    )

    tags: List[str] = field(
        metadata={
            'metadata': {
                'description': 'List of tags associated with this instance',
                'example': '["tag1", "tag2"]',
            }
        }
    )

    image: str = field(
        metadata={
            'marshmallow_field': FieldWithId(),
            'metadata': {
                'description': 'Image used to ',
                'example': 'linode/archlinux2014.04',
            },
        }
    )

    region: str = field(
        metadata={
            'marshmallow_field': FieldWithId(),
            'metadata': {
                'description': 'Region where the instance is located',
                'example': 'eu-west',
            },
        }
    )

    hypervisor: str = field(
        metadata={
            'metadata': {
                'description': 'The virtualization engine powering this instance',
                'example': 'kvm',
            }
        }
    )

    specs: LinodeInstanceSpecs
    alerts: LinodeInstanceAlerts
    backups: LinodeInstanceBackups

    created_at: datetime = field(
        metadata={
            'data_key': 'created',
            'metadata': {
                'description': 'Instance creation date',
                'example': '2020-01-01T00:00:00Z',
            },
        }
    )

    updated_at: datetime = field(
        metadata={
            'data_key': 'updated',
            'metadata': {
                'description': 'When the instance was last polled/updated',
                'example': '2020-01-01T01:00:00Z',
            },
        }
    )


LinodeInstanceSchema = class_schema(LinodeInstance, base_schema=LinodeBaseSchema)



from marshmallow import fields
from marshmallow.schema import Schema

from platypush.schemas import StrippedString


class TensorflowTrainSchema(Schema):
    """
    Schema for TensorFlow model training results.
    """

    model = StrippedString(
        required=True,
        metadata={
            "description": "Model name.",
            "example": "MyModel",
        },
    )

    epochs = fields.Int(
        required=True,
        metadata={
            "description": "Number of epochs.",
            "example": 10,
        },
    )

    history = fields.Dict(
        metadata={
            "description": "Training history.",
            "example": {
                "loss": [0.1, 0.2, 0.3],
                "accuracy": [0.9, 0.8, 0.7],
            },
        },
    )



from platypush.message.event import Event


class MailEvent(Event):
    """
    Base class for mail events.
    """

    def __init__(self, *args, account: str, folder: str, message, **kwargs):
        super().__init__(
            *args, account=account, folder=folder, message=message, **kwargs
        )


class UnseenMailEvent(MailEvent):
    """
    Triggered when a new email is received or marked as unseen.
    """


class SeenMailEvent(MailEvent):
    """
    Triggered when a previously unseen email is seen.
    """


class FlaggedMailEvent(MailEvent):
    """
    Triggered when a message is marked as flagged/starred.
    """


class UnflaggedMailEvent(MailEvent):
    """
    Triggered when a message previously marked as flagged/starred is unflagged.
    """


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
import datetime
from enum import Enum
import json
import os
import re
import subprocess
import tempfile
import textwrap
from typing import Iterable, Optional, Union
from urllib.parse import urlparse

from platypush.plugins import Plugin, action


@dataclass
class OutputFormat:
    """
    Definition of a supported output format.
    """

    name: str
    cmd_fmt: str
    extensions: Iterable[str] = ()


class OutputFormats(Enum):
    """
    Supported output formats.
    """

    HTML = OutputFormat('html', extensions=('html', 'htm'), cmd_fmt='html')
    # PDF will first be exported to HTML and then converted to PDF
    PDF = OutputFormat('pdf', extensions=('pdf',), cmd_fmt='html')
    TEXT = OutputFormat('text', extensions=('txt',), cmd_fmt='text')
    MARKDOWN = OutputFormat('markdown', extensions=('md',), cmd_fmt='markdown')

    @classmethod
    def parse(
        cls,
        type: Union[str, "OutputFormats"],  # pylint: disable=redefined-builtin
        outfile: Optional[str] = None,
    ) -> "OutputFormats":
        """
        Parse the format given a type argument and and output file name.
        """
        try:
            fmt = (
                getattr(OutputFormats, type.upper()) if isinstance(type, str) else type
            )
        except AttributeError as e:
            raise AssertionError(
                f'Unsupported output format: {type}. Supported formats: '
                + f'{[f.name for f in OutputFormats]}'
            ) from e

        by_extension = {ext.lower(): f for f in cls for ext in f.value.extensions}
        if outfile:
            fmt_by_ext = by_extension.get(os.path.splitext(outfile)[1].lower()[1:])
            if fmt_by_ext:
                return fmt_by_ext

        return fmt


class HttpWebpagePlugin(Plugin):
    """
    Plugin to handle and parse/simplify web pages.
    It used to use the Mercury Reader web API, but now that the API is discontinued this plugin is basically a
    wrapper around the `mercury-parser <https://github.com/postlight/mercury-parser>`_ JavaScript library.

    Requires:

        * The mercury-parser library installed (``npm install -g @postlight/mercury-parser``)

    """

    _mercury_script = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'mercury-parser.js'
    )

    @staticmethod
    def _parse(proc):
        """
        Runs the mercury-parser script and returns the result as a string.
        """
        with subprocess.Popen(proc, stdout=subprocess.PIPE, stderr=None) as parser:
            return parser.communicate()[0].decode()

    @staticmethod
    def _fix_relative_links(markdown: str, url: str) -> str:
        """
        Fix relative links to match the base URL of the page (Markdown only).
        """
        parsed_url = urlparse(url)
        base_url = f'{parsed_url.scheme}://{parsed_url.netloc}'
        return re.sub(r'(\[.+?])\((/.+?)\)', fr'\1({base_url}\2)', markdown)

    @action
    def simplify(
        self,
        url: str,
        type: Union[  # pylint: disable=redefined-builtin
            str, OutputFormats
        ] = OutputFormats.HTML,
        html: Optional[str] = None,
        outfile: Optional[str] = None,
        font_size: str = '19px',
        font_family: Union[str, Iterable[str]] = (
            '-apple-system',
            'Segoe UI',
            'Roboto',
            'Oxygen',
            'Ubuntu',
            'Cantarell',
            "Fira Sans",
            'Open Sans',
            'Droid Sans',
            'Helvetica Neue',
            'Helvetica',
            'Arial',
            'sans-serif',
        ),
    ):
        """
        Parse the readable content of a web page removing any extra HTML elements using Mercury.

        :param url: URL to parse.
        :param type: Output format. Supported types: ``html``, ``markdown``,
            ``text``, ``pdf`` (default: ``html``).
        :param html: Set this parameter if you want to parse some HTML content
            already fetched. Note that URL is still required by Mercury to
            properly style the output, but it won't be used to actually fetch
            the content.
        :param outfile: If set then the output will be written to the specified
            file. If the file extension is ``.pdf`` then the content will be
            exported in PDF format. If the output ``type`` is not specified
            then it can also be inferred from the extension of the output file.
        :param font_size: Font size to use for the output (default: 19px).
        :param font_family: Custom font family (or list of font families, in
            decreasing order) to use for the output. It only applies to HTML
            and PDF.
        :return: dict

        Example return payload outfile is not specified::

            {
                "url": <url>,
                "title": <page title>,
                "content": <page parsed content>
            }

        Example return payload if outfile is specified::

            {
                "url": <url>,
                "title": <page title>,
                "outfile": <output file absolute path>
            }

        """

        self.logger.info('Parsing URL %s', url)
        fmt = OutputFormats.parse(type=type, outfile=outfile)
        proc = ['node', self._mercury_script, url, fmt.value.cmd_fmt]
        tmp_file = None

        if html:
            with tempfile.NamedTemporaryFile('w+', delete=False) as f:
                tmp_file = f.name
                f.write(html)
                f.flush()
                proc.append(f.name)

        try:
            response = self._parse(proc)
        finally:
            if tmp_file:
                os.unlink(tmp_file)

        try:
            response = json.loads(response.strip())
        except Exception as e:
            raise RuntimeError(
                f'Could not parse JSON: {e}. Response: {response}'
            ) from e

        if fmt == OutputFormats.MARKDOWN:
            response['content'] = self._fix_relative_links(response['content'], url)

        self.logger.debug('Got response from Mercury API: %s', response)
        title = response.get(
            'title',
            (
                ('Published' if response.get('date_published') else 'Generated')
                + ' on '
                + (
                    response.get('date_published')
                    or datetime.datetime.now().isoformat()
                )
            ),
        )

        content = response.get('content', '')

        if not outfile:
            return {
                'url': url,
                'title': title,
                'content': content,
            }

        return self._process_outfile(
            url=url,
            fmt=fmt,
            title=title,
            content=content,
            outfile=outfile,
            font_size=font_size,
            font_family=tuple(
                font_family,
            )
            if isinstance(font_family, str)
            else tuple(font_family),
        )

    @staticmethod
    def _style_by_format(
        fmt: OutputFormats,
        font_size: str,
        font_family: Iterable[str],
    ) -> str:
        """
        :return: The CSS style to be used for the given output format.
        """
        style = textwrap.dedent(
            f'''
            ._parsed-content-container {{
                font-size: {font_size};
                font-family: {', '.join(f'"{f}"' for f in font_family)};
            }}

            ._parsed-content {{
                text-align: justify;
            }}

            pre {{
                white-space: pre-wrap;
            }}
            '''
        )

        if fmt == OutputFormats.HTML:
            style += textwrap.dedent(
                '''
                ._parsed-content-container {
                    margin: 1em;
                    display: flex;
                    flex-direction: column;
                    align-items: center;
                }

                ._parsed-content {
                    max-width: 800px;
                }

                h1 {
                    max-width: 800px;
                }
                '''
            )

        return style

    @classmethod
    def _process_outfile(
        cls,
        url: str,
        fmt: OutputFormats,
        title: str,
        content: str,
        outfile: str,
        font_size: str,
        font_family: Iterable[str],
    ):
        """
        Process the output file.

        :param url: URL to parse.
        :param fmt: Output format. Supported types: ``html``, ``markdown``,
            ``text``, ``pdf`` (default: ``html``).
        :param title: Page title.
        :param content: Page content.
        :param outfile: Output file path.
        :param font_size: Font size to use for the output (default: 19px).
        :param font_family: Custom font family (or list of font families, in
            decreasing order) to use for the output. It only applies to HTML
            and PDF.
        :return: dict
        """
        outfile = os.path.abspath(os.path.expanduser(outfile))
        style = cls._style_by_format(fmt, font_size, font_family)

        if fmt in {OutputFormats.HTML, OutputFormats.PDF}:
            content = textwrap.dedent(
                f'''
                <div class="_parsed-content-container">
                    <h1><a href="{url}" target="_blank">{title}</a></h1>
                    <div class="_parsed-content">{content}</div>
                </div>
                '''
            )

            if fmt == OutputFormats.PDF:
                content = textwrap.dedent(
                    f'''<html>
                            <head>
                                <style>{style}</style>
                                <title>{title}</title>
                            </head>
                            <body>
                              {content}
                            </body>
                        </html>
                    '''
                )
            else:
                content = textwrap.dedent(
                    f'''
                    <style>
                        {style}
                    </style>
                    {content}
                    '''
                )
        elif fmt == OutputFormats.MARKDOWN:
            content = f'# [{title}]({url})\n\n{content}'

        if fmt == OutputFormats.PDF:
            cls._process_pdf(content, outfile, style)
        else:
            with open(outfile, 'w', encoding='utf-8') as f:
                f.write(content)

        return {
            'url': url,
            'title': title,
            'outfile': outfile,
        }

    @staticmethod
    def _process_pdf(content: str, outfile: str, style: str):
        """
        Convert the given HTML content to a PDF document.

        :param content: Page content.
        :param outfile: Output file path.
        :param style: CSS style to use for the output.
        """
        import weasyprint

        try:
            from weasyprint.fonts import FontConfiguration  # pylint: disable
        except ImportError:
            from weasyprint.document import FontConfiguration

        font_config = FontConfiguration()
        css = [
            weasyprint.CSS(string=style, font_config=font_config),
        ]

        weasyprint.HTML(string=content).write_pdf(outfile, stylesheets=css)


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field
from typing import Optional

from platypush.schemas.dataclasses import percent_field


@dataclass
class Battery:
    """
    System battery sensor wrapper.
    """

    seconds_left: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'High threshold for the temperature sensor, in Celsius',
                'example': 75,
            }
        }
    )

    power_plugged: Optional[bool] = field(
        metadata={
            'metadata': {
                'description': 'Whether the battery is plugged in or not',
                'example': False,
            }
        }
    )

    value: Optional[float] = percent_field(
        metadata={
            'metadata': {
                'description': 'Current charge left, as a percentage value '
                'between 0 and 1',
                'example': 0.5,
            }
        }
    )



from abc import ABC
from typing import Any, List, Optional

from platypush.message.event import Event


class MidiEvent(Event, ABC):
    """
    Base class for MIDI events.
    """

    def __init__(self, *args, device: Optional[str], port: Optional[int], **kwargs):
        """
        :param device: The MIDI device name.
        :param port: The MIDI device port number.
        """
        super().__init__(*args, device=device, port=port, **kwargs)


class MidiMessageEvent(MidiEvent):
    """
    Event triggered upon received MIDI message.
    """

    def __init__(self, *args, message: List[int], data: Optional[Any] = None, **kwargs):
        """
        :param message: The received MIDI message.
        :param data: Additional data associated to the event.
        """
        super().__init__(*args, message=message, data=data, **kwargs)


class MidiDeviceConnectedEvent(MidiEvent):
    """
    Event triggered when a MIDI device is connected.
    """


class MidiDeviceDisconnectedEvent(MidiEvent):
    """
    Event triggered when a MIDI device is disconnected.
    """


class MidiDeviceAddedEvent(MidiEvent):
    """
    Event triggered when a MIDI device is added to the list of available devices.
    """


class MidiDeviceRemovedEvent(MidiEvent):
    """
    Event triggered when a MIDI device is removed from the list of available devices.
    """


# vim:sw=4:ts=4:et:
